from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from cognite.powerops.clients.data_classes._core import DomainModel, DomainModelApply, TypeList

if TYPE_CHECKING:
    from cognite.powerops.clients.data_classes._output_containers import OutputContainerApply
    from cognite.powerops.clients.data_classes._scenario_mappings import ScenarioMappingApply

__all__ = ["ScenarioTemplate", "ScenarioTemplateApply", "ScenarioTemplateList"]


class ScenarioTemplate(DomainModel):
    space: ClassVar[str] = "power-ops"
    base_mapping: Optional[str] = Field(None, alias="baseMapping")
    model: Optional[str] = None
    output_definitions: Optional[str] = Field(None, alias="outputDefinitions")
    shop_files: list[str] = Field([], alias="shopFiles")
    shop_version: Optional[str] = Field(None, alias="shopVersion")
    template_version: Optional[str] = Field(None, alias="templateVersion")
    watercourse: Optional[str] = None


class ScenarioTemplateApply(DomainModelApply):
    space: ClassVar[str] = "power-ops"
    base_mapping: Optional[Union["ScenarioMappingApply", str]] = Field(None, repr=False)
    model: Optional[str] = None
    output_definitions: Optional[Union["OutputContainerApply", str]] = Field(None, repr=False)
    shop_files: list[str] = []
    shop_version: Optional[str] = None
    template_version: Optional[str] = None
    watercourse: Optional[str] = None

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        source = dm.NodeOrEdgeData(
            source=dm.ContainerId("power-ops", "ScenarioTemplate"),
            properties={
                "baseMapping": {
                    "space": "power-ops",
                    "externalId": self.base_mapping
                    if isinstance(self.base_mapping, str)
                    else self.base_mapping.external_id,
                },
                "model": self.model,
                "outputDefinitions": {
                    "space": "power-ops",
                    "externalId": self.output_definitions
                    if isinstance(self.output_definitions, str)
                    else self.output_definitions.external_id,
                },
                "shopFiles": self.shop_files,
                "shopVersion": self.shop_version,
                "templateVersion": self.template_version,
                "watercourse": self.watercourse,
            },
        )
        sources.append(source)

        this_node = dm.NodeApply(
            space=self.space,
            external_id=self.external_id,
            existing_version=self.existing_version,
            sources=sources,
        )
        nodes = [this_node]
        edges = []
        cache.add(self.external_id)

        if isinstance(self.base_mapping, DomainModelApply):
            instances = self.base_mapping._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        if isinstance(self.output_definitions, DomainModelApply):
            instances = self.output_definitions._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))


class ScenarioTemplateList(TypeList[ScenarioTemplate]):
    _NODE = ScenarioTemplate



from __future__ import annotations

from pathlib import Path
from typing import ClassVar, Literal

from pydantic import BaseModel, ConfigDict, Field


class ShopFileConfig(BaseModel):
    model_config: ClassVar[ConfigDict] = ConfigDict(populate_by_name=True)
    watercourse_name: str
    # Alias for backwards compatibility
    file_path: Path = Field(alias="path")
    cogshop_file_type: Literal[
        "case",
        "model",
        "log",
        "commands",
        "extra_data",
        "water_value_cut_file_reservoir_mapping",
        "water_value_cut_file",
        "module_series",
    ]

    @property
    def external_id(self) -> str:
        return f"SHOP_{self.watercourse_name}_{self.file_path.stem}"



from __future__ import annotations

import json
from typing import Dict, List, Optional, Union

from pydantic import BaseModel, ConfigDict, validator

from cognite.powerops.resync.config.shared import (
    AggregationMethod,
    RetrievalType,
    TimeSeriesMapping,
    TimeSeriesMappingEntry,
    Transformation,
    TransformationType,
)


class Configuration(BaseModel):
    model_config = ConfigDict(populate_by_name=True)


class PriceScenarioID(BaseModel):
    id: str
    rename: str = ""


class PriceScenario(BaseModel):
    name: str
    time_series_external_id: Optional[str] = None
    transformations: Optional[List[Transformation]] = None

    def to_time_series_mapping(self) -> TimeSeriesMapping:
        retrieve = RetrievalType.RANGE if self.time_series_external_id else None
        transformations = self.transformations or []

        # to make buy price slightly higher than sale price in SHOP
        transformations_buy_price = transformations + [
            Transformation(transformation=TransformationType.ADD, kwargs={"value": 0.01})
        ]

        sale_price_row = TimeSeriesMappingEntry(
            object_type="market",
            object_name=self.name,
            attribute_name="sale_price",
            time_series_external_id=self.time_series_external_id,
            transformations=transformations,
            retrieve=retrieve,
            aggregation=AggregationMethod.mean,
        )

        buy_price_row = TimeSeriesMappingEntry(
            object_type="market",
            object_name=self.name,
            attribute_name="buy_price",
            time_series_external_id=self.time_series_external_id,
            transformations=transformations_buy_price,
            retrieve=retrieve,
            aggregation=AggregationMethod.mean,
        )

        return TimeSeriesMapping(rows=[sale_price_row, buy_price_row])


class RelativeTime(BaseModel):
    relative_time_string: Optional[str] = None
    operations: Optional[list[tuple[str, Union[str, Dict[str, int]]]]] = None

    @validator("operations", pre=True, always=True)
    def to_old_format(cls, value):
        if not isinstance(value, list):
            return value

        old_formats = []
        for v in value:
            if isinstance(v, dict):
                operation, argument = next(iter(v.items()))
                old_formats.append((operation, argument))
            else:
                # Already old format
                old_formats.append(v)
        return old_formats

    @validator("operations", pre=True, always=True)
    def _parse_relative_time_string(cls, operations, values):
        # NOTE: tuples will be parsed as lists when dumping to string
        if operations:
            return operations
        elif values["relative_time_string"] == "tomorrow":
            return [("shift", {"days": 1}), ("floor", "day")]
        elif values["relative_time_string"] == "end_of_next_week":
            return [("floor", "week"), ("shift", {"weeks": 2})]
        elif values["relative_time_string"] == "monday":
            # Monday next week (given that we are before friday 12:00)
            return [("shift", {"hours": 12}), ("shift", {"weekday": 4}), ("floor", "day"), ("shift", {"weekday": 0})]
        elif values["relative_time_string"] == "saturday":
            # This Saturday (given that we are before thursday 12:00)
            return [("shift", {"hours": 12}), ("shift", {"weekday": 3}), ("floor", "day"), ("shift", {"weekday": 5})]
        else:
            raise ValueError(f"{values['relative_time_string']} not a valid value for relative_time_string")

    def __str__(self) -> str:
        return json.dumps(self.operations)



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from cognite.powerops.clients.data_classes._core import DomainModel, DomainModelApply, TypeList

if TYPE_CHECKING:
    from cognite.powerops.clients.data_classes._value_transformations import ValueTransformationApply

__all__ = ["InputTimeSeriesMapping", "InputTimeSeriesMappingApply", "InputTimeSeriesMappingList"]


class InputTimeSeriesMapping(DomainModel):
    space: ClassVar[str] = "power-ops"
    aggregation: Optional[str] = None
    cdf_time_series: Optional[str] = Field(None, alias="cdfTimeSeries")
    retrieve: Optional[str] = None
    shop_attribute_name: Optional[str] = Field(None, alias="shopAttributeName")
    shop_object_name: Optional[str] = Field(None, alias="shopObjectName")
    shop_object_type: Optional[str] = Field(None, alias="shopObjectType")
    transformations: list[str] = []


class InputTimeSeriesMappingApply(DomainModelApply):
    space: ClassVar[str] = "power-ops"
    aggregation: Optional[str] = None
    cdf_time_series: Optional[str] = None
    retrieve: Optional[str] = None
    shop_attribute_name: Optional[str] = None
    shop_object_name: Optional[str] = None
    shop_object_type: Optional[str] = None
    transformations: list[Union["ValueTransformationApply", str]] = Field(default_factory=list, repr=False)

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        source = dm.NodeOrEdgeData(
            source=dm.ContainerId("power-ops", "InputTimeSeriesMapping"),
            properties={
                "aggregation": self.aggregation,
                "cdfTimeSeries": self.cdf_time_series,
                "retrieve": self.retrieve,
                "shopAttributeName": self.shop_attribute_name,
                "shopObjectName": self.shop_object_name,
                "shopObjectType": self.shop_object_type,
            },
        )
        sources.append(source)

        this_node = dm.NodeApply(
            space=self.space,
            external_id=self.external_id,
            existing_version=self.existing_version,
            sources=sources,
        )
        nodes = [this_node]
        edges = []
        cache.add(self.external_id)

        for transformation in self.transformations:
            edge = self._create_transformation_edge(transformation)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(transformation, DomainModelApply):
                instances = transformation._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))

    def _create_transformation_edge(self, transformation: Union[str, "ValueTransformationApply"]) -> dm.EdgeApply:
        if isinstance(transformation, str):
            end_node_ext_id = transformation
        elif isinstance(transformation, DomainModelApply):
            end_node_ext_id = transformation.external_id
        else:
            raise TypeError(f"Expected str or ValueTransformationApply, got {type(transformation)}")

        return dm.EdgeApply(
            space="power-ops",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("power-ops", "InputTimeSeriesMapping.transformations"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("power-ops", end_node_ext_id),
        )


class InputTimeSeriesMappingList(TypeList[InputTimeSeriesMapping]):
    _NODE = InputTimeSeriesMapping



from __future__ import annotations

import logging
from pathlib import Path
from typing import Annotated, Optional

import typer
from rich.console import Console
from rich.logging import RichHandler
from rich.pretty import pprint

from cognite import powerops
from cognite.powerops import resync
from cognite.powerops._models import MODEL_BY_NAME
from cognite.powerops.clients.powerops_client import get_powerops_client

logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)
logging.getLogger("msal").setLevel(logging.WARNING)
logging.getLogger("cognite-sdk").setLevel(logging.WARNING)

FORMAT = "%(message)s"
logging.basicConfig(level="NOTSET", format=FORMAT, datefmt="[%X]", handlers=[RichHandler()])

log = logging.getLogger("rich")

app = typer.Typer(pretty_exceptions_short=False, pretty_exceptions_show_locals=False, pretty_exceptions_enable=False)


def _version_callback(value: bool):
    if value:
        typer.echo(powerops.__version__)
        raise typer.Exit()


@app.callback()
def common(
    ctx: typer.Context,
    version: bool = typer.Option(None, "--version", callback=_version_callback),
):
    ...


@app.command(
    "plan",
    help="Preview the changes from the configuration files that `powerops apply` would make to the data model in CDF",
)
def plan(
    path: Annotated[Path, typer.Argument(help="Path to configuration files")],
    market: Annotated[str, typer.Argument(help="Selected power market")],
    models: list[str] = typer.Option(
        default=sorted(resync.DEFAULT_MODELS),
        help=f"The models to run the plan. Available models: {', '.join(resync.MODEL_BY_NAME)}",
    ),
    dump_folder: Optional[Path] = typer.Option(
        default=None, help="If present, the local and cdf changes will be dumped to this directory."
    ),
):
    if dump_folder and not dump_folder.is_dir():
        raise typer.BadParameter(f"{dump_folder} is not a directory")

    log.info(f"Running plan on configuration files located in {path}")
    if len(models) == 1 and models[0].lower() == "all":
        models = list(MODEL_BY_NAME.keys())

    resync.plan(path, market, echo=log.info, model_names=models, dump_folder=dump_folder, echo_pretty=pprint)


@app.command("apply", help="Apply the changes from the configuration files to the data model in CDF")
def apply(
    path: Annotated[Path, typer.Argument(help="Path to configuration files")],
    market: Annotated[str, typer.Argument(help="Selected power market")],
    models: list[str] = typer.Option(
        default=sorted(resync.DEFAULT_MODELS),
        help=f"The models to run apply. Available models: {', '.join(resync.MODEL_BY_NAME)}",
    ),
    auto_yes: bool = typer.Option(False, "--yes", "-y", help="Auto confirm all prompts"),
):
    log.info(f"Running apply on configuration files located in {path}")

    resync.apply(path, market, model_names=models, echo=log.info, auto_yes=auto_yes, echo_pretty=pprint)


@app.command(
    "deploy",
    help=f"Deploy the data model in CDF. Available models: {list(MODEL_BY_NAME.keys())}. "
    f"Use 'all' to deploy all models.",
)
def deploy(
    models: Annotated[list[str], typer.Argument(help="The models to deploy")],
):
    if len(models) == 1 and models[0].lower() == "all":
        models = list(MODEL_BY_NAME.keys())

    client = get_powerops_client()
    for model_name in models:
        if model_name not in MODEL_BY_NAME:
            log.warning(f"Model {model_name} not found, skipping. Available models: {list(MODEL_BY_NAME.keys())}")
            continue
        log.info(f"Deploying {model_name} model...")

        model = MODEL_BY_NAME[model_name]
        result = client.cdf.data_modeling.graphql.apply_dml(
            model.id_,
            model.graphql,
            model.name,
            model.description,
        )
        log.info(f"Deployed {model_name} model ({result.space}, {result.external_id}, {result.version})")


@app.command("show", help=f"Show the graphql schema of Power Ops model. Available models: {list(MODEL_BY_NAME.keys())}")
def show(
    model: Annotated[str, typer.Argument(help="The models to deploy")],
    remove_newlines: bool = typer.Option(
        False,
        "--remove-newlines",
        help="Remove newlines from the graphql schema. This is done when deploying the schema.",
    ),
):
    if model not in MODEL_BY_NAME:
        log.warning(f"Model {model} not found. Available models: {list(MODEL_BY_NAME.keys())}")
        typer.Exit()

    console = Console()
    graphql = MODEL_BY_NAME[model].graphql
    if remove_newlines:
        graphql = graphql.replace("\n", "")
        console.print(f"{graphql!r}")
    else:
        console.print(graphql)


def main():
    app()


if __name__ == "__main__":
    deploy(["dayahead"])



from __future__ import annotations

from collections import defaultdict
from typing import Dict, List, Sequence, Tuple, overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client._constants import INSTANCES_LIST_LIMIT_DEFAULT

from cognite.powerops.clients._api._core import TypeAPI
from cognite.powerops.clients.data_classes import DayAheadProces, DayAheadProcesApply, DayAheadProcesList


class DayAheadProcesBidMatrixGeneratorConfigsAPI:
    def __init__(self, client: CogniteClient):
        self._client = client

    def retrieve(self, external_id: str | Sequence[str]) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "DayAheadProcess.bidMatrixGeneratorConfig"},
        )
        if isinstance(external_id, str):
            is_day_ahead_proces = f.Equals(
                ["edge", "startNode"],
                {"space": "power-ops", "externalId": external_id},
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_day_ahead_proces)
            )

        else:
            is_day_ahead_process = f.In(
                ["edge", "startNode"],
                [{"space": "power-ops", "externalId": ext_id} for ext_id in external_id],
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_day_ahead_process)
            )

    def list(self, limit=INSTANCES_LIST_LIMIT_DEFAULT) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "DayAheadProcess.bidMatrixGeneratorConfig"},
        )
        return self._client.data_modeling.instances.list("edge", limit=limit, filter=is_edge_type)


class DayAheadProcesIncrementalMappingsAPI:
    def __init__(self, client: CogniteClient):
        self._client = client

    def retrieve(self, external_id: str | Sequence[str]) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "DayAheadProcess.incremental_mappings"},
        )
        if isinstance(external_id, str):
            is_day_ahead_proces = f.Equals(
                ["edge", "startNode"],
                {"space": "power-ops", "externalId": external_id},
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_day_ahead_proces)
            )

        else:
            is_day_ahead_process = f.In(
                ["edge", "startNode"],
                [{"space": "power-ops", "externalId": ext_id} for ext_id in external_id],
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_day_ahead_process)
            )

    def list(self, limit=INSTANCES_LIST_LIMIT_DEFAULT) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "DayAheadProcess.incremental_mappings"},
        )
        return self._client.data_modeling.instances.list("edge", limit=limit, filter=is_edge_type)


class DayAheadProcessAPI(TypeAPI[DayAheadProces, DayAheadProcesApply, DayAheadProcesList]):
    def __init__(self, client: CogniteClient):
        super().__init__(
            client=client,
            sources=dm.ViewId("power-ops", "DayAheadProcess", "dd1bf62feefc9a"),
            class_type=DayAheadProces,
            class_apply_type=DayAheadProcesApply,
            class_list=DayAheadProcesList,
        )
        self.bid_matrix_generator_configs = DayAheadProcesBidMatrixGeneratorConfigsAPI(client)
        self.incremental_mappings = DayAheadProcesIncrementalMappingsAPI(client)

    def apply(self, day_ahead_proces: DayAheadProcesApply, replace: bool = False) -> dm.InstancesApplyResult:
        instances = day_ahead_proces.to_instances_apply()
        return self._client.data_modeling.instances.apply(nodes=instances.nodes, edges=instances.edges, replace=replace)

    def delete(self, external_id: str | Sequence[str]) -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(DayAheadProcesApply.space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(DayAheadProcesApply.space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> DayAheadProces:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> DayAheadProcesList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> DayAheadProces | DayAheadProcesList:
        if isinstance(external_id, str):
            day_ahead_proces = self._retrieve((self.sources.space, external_id))

            bid_matrix_generator_config_edges = self.bid_matrix_generator_configs.retrieve(external_id)
            day_ahead_proces.bid_matrix_generator_config = [
                edge.end_node.external_id for edge in bid_matrix_generator_config_edges
            ]
            incremental_mapping_edges = self.incremental_mappings.retrieve(external_id)
            day_ahead_proces.incremental_mappings = [edge.end_node.external_id for edge in incremental_mapping_edges]

            return day_ahead_proces
        else:
            day_ahead_process = self._retrieve([(self.sources.space, ext_id) for ext_id in external_id])

            bid_matrix_generator_config_edges = self.bid_matrix_generator_configs.retrieve(external_id)
            self._set_bid_matrix_generator_config(day_ahead_process, bid_matrix_generator_config_edges)
            incremental_mapping_edges = self.incremental_mappings.retrieve(external_id)
            self._set_incremental_mappings(day_ahead_process, incremental_mapping_edges)

            return day_ahead_process

    def list(self, limit: int = INSTANCES_LIST_LIMIT_DEFAULT) -> DayAheadProcesList:
        day_ahead_process = self._list(limit=limit)

        bid_matrix_generator_config_edges = self.bid_matrix_generator_configs.list(limit=-1)
        self._set_bid_matrix_generator_config(day_ahead_process, bid_matrix_generator_config_edges)
        incremental_mapping_edges = self.incremental_mappings.list(limit=-1)
        self._set_incremental_mappings(day_ahead_process, incremental_mapping_edges)

        return day_ahead_process

    @staticmethod
    def _set_bid_matrix_generator_config(
        day_ahead_process: Sequence[DayAheadProces], bid_matrix_generator_config_edges: Sequence[dm.Edge]
    ):
        edges_by_start_node: Dict[Tuple, List] = defaultdict(list)
        for edge in bid_matrix_generator_config_edges:
            edges_by_start_node[edge.start_node.as_tuple()].append(edge)

        for day_ahead_proces in day_ahead_process:
            node_id = day_ahead_proces.id_tuple()
            if node_id in edges_by_start_node:
                day_ahead_proces.bid_matrix_generator_config = [
                    edge.end_node.external_id for edge in edges_by_start_node[node_id]
                ]

    @staticmethod
    def _set_incremental_mappings(
        day_ahead_process: Sequence[DayAheadProces], incremental_mapping_edges: Sequence[dm.Edge]
    ):
        edges_by_start_node: Dict[Tuple, List] = defaultdict(list)
        for edge in incremental_mapping_edges:
            edges_by_start_node[edge.start_node.as_tuple()].append(edge)

        for day_ahead_proces in day_ahead_process:
            node_id = day_ahead_proces.id_tuple()
            if node_id in edges_by_start_node:
                day_ahead_proces.incremental_mappings = [
                    edge.end_node.external_id for edge in edges_by_start_node[node_id]
                ]



from __future__ import annotations

from cognite.client import ClientConfig, CogniteClient

from cognite.powerops.clients._api_client import (
    BenchmarkAPIs,
    CogShopAPIs,
    DayAheadAPIs,
    ProductionAPIs,
    RKOMMarketAPIs,
)
from cognite.powerops.clients.data_set_api import DataSetsAPI
from cognite.powerops.clients.shop import ShopClient
from cognite.powerops.utils.cdf import Settings, get_client_config
from cognite.powerops.cogshop1 import CogShop1Client


class PowerOpsClient:
    def __init__(
        self,
        read_dataset: str,
        write_dataset: str,
        cogshop_version: str,
        config: ClientConfig,
    ):
        self.cdf = CogniteClient(config)
        data_set_api = DataSetsAPI(self.cdf, read_dataset, write_dataset)

        self.production = ProductionAPIs(self.cdf)
        self.dayahead = DayAheadAPIs(self.cdf)
        self.rkom = RKOMMarketAPIs(self.cdf)
        self.benchmark = BenchmarkAPIs(self.cdf)
        self.cog_shop = CogShopAPIs(self.cdf)
        self.shop = ShopClient(self.cdf, self.cog_shop, data_set_api, cogshop_version)
        self.cog_shop1 = CogShop1Client(self.cdf)


def get_powerops_client(write_dataset: str | None = None) -> PowerOpsClient:
    settings = Settings(**{"powerops": {"write_dataset": write_dataset}})
    client_config = get_client_config(settings.cognite)

    return PowerOpsClient(
        settings.powerops.read_dataset,
        settings.powerops.write_dataset,
        settings.powerops.cogshop_version,
        config=client_config,
    )



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from cognite.powerops.cogshop1.data_classes._core import DomainModel, DomainModelApply, TypeList

if TYPE_CHECKING:
    from cognite.powerops.cogshop1.data_classes._file_refs import FileRefApply
    from cognite.powerops.cogshop1.data_classes._mappings import MappingApply

__all__ = ["ModelTemplate", "ModelTemplateApply", "ModelTemplateList"]


class ModelTemplate(DomainModel):
    space: ClassVar[str] = "cogShop"
    base_mappings: list[str] = Field([], alias="baseMappings")
    model: Optional[str] = None
    shop_version: Optional[str] = Field(None, alias="shopVersion")
    version: Optional[str] = None
    watercourse: Optional[str] = None


class ModelTemplateApply(DomainModelApply):
    space: ClassVar[str] = "cogShop"
    base_mappings: list[Union["MappingApply", str]] = Field(default_factory=list, repr=False)
    model: Optional[Union["FileRefApply", str]] = Field(None, repr=False)
    shop_version: str
    version: str
    watercourse: str

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        source = dm.NodeOrEdgeData(
            source=dm.ContainerId("cogShop", "ModelTemplate"),
            properties={
                "model": {
                    "space": "cogShop",
                    "externalId": self.model if isinstance(self.model, str) else self.model.external_id,
                },
                "shopVersion": self.shop_version,
                "version": self.version,
                "watercourse": self.watercourse,
            },
        )
        sources.append(source)

        this_node = dm.NodeApply(
            space=self.space,
            external_id=self.external_id,
            existing_version=self.existing_version,
            sources=sources,
        )
        nodes = [this_node]
        edges = []
        cache.add(self.external_id)

        for base_mapping in self.base_mappings:
            edge = self._create_base_mapping_edge(base_mapping)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(base_mapping, DomainModelApply):
                instances = base_mapping._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        if isinstance(self.model, DomainModelApply):
            instances = self.model._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))

    def _create_base_mapping_edge(self, base_mapping: Union[str, "MappingApply"]) -> dm.EdgeApply:
        if isinstance(base_mapping, str):
            end_node_ext_id = base_mapping
        elif isinstance(base_mapping, DomainModelApply):
            end_node_ext_id = base_mapping.external_id
        else:
            raise TypeError(f"Expected str or MappingApply, got {type(base_mapping)}")

        return dm.EdgeApply(
            space="cogShop",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("cogShop", "ModelTemplate.baseMappings"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("cogShop", end_node_ext_id),
        )


class ModelTemplateList(TypeList[ModelTemplate]):
    _NODE = ModelTemplate



import doctest
import operator
from collections.abc import Iterable

from functools import reduce
from pprint import pformat
from deepdiff.model import PrettyOrderedSet
from typing import Any, Union, Type, get_args, get_origin
from types import GenericAlias


from cognite.client.data_classes import Relationship


def isinstance_list(value: Any, type_: Type):
    return isinstance(value, list) and value and isinstance(value[0], type_)


def match_field_from_relationship(model_fields: Iterable[str], relationship: Relationship) -> str:
    """Find the field on the model that matches the relationship using the label"""
    if len(relationship.labels) != 1:
        raise ValueError(f"Expected one label in {relationship.labels=}")
    label = relationship.labels[0].external_id.split(".")[-1]

    # Market labels sometimes have a suffix of _{target_type}
    candidates = [
        field for field in model_fields if label in field or label.removesuffix(f"_{relationship.target_type}") in field
    ]

    if len(candidates) != 1:
        raise ValueError(f"Could not match {relationship.external_id=} to {model_fields=}")

    return candidates[0]


def pydantic_model_class_candidate(class_: type = None) -> bool:
    """GenericAlias is a potential list, get origin checks an optional field"""
    return isinstance(class_, GenericAlias) or (get_origin(class_) is Union and type(None) in get_args(class_))


def format_change_binary(
    deep_diff: dict[str, dict],
) -> list[str]:
    """
    Formats a dict of changes with updated values to a list of strings

    >>> deep_diff = {
    ...     "root[0][plants][name]": {
    ...         "old_value": "Old name",
    ...         "new_value": "New name",
    ...     },
    ... }
    >>> format_change_binary(deep_diff)
    [' * [0][plants][name]:\\n', "\\t- 'Old name'\\t", '  -->   ', "'New name'", '\\n']
    """
    str_builder = []
    for path_, change_dict in deep_diff.items():
        str_builder.extend(
            (
                f" * {path_.replace('root', '') }:\n",
                f'\t- {pformat(change_dict.get("old_value"))}\t',
                "  -->   ",
                f'{pformat(change_dict.get("new_value"))}',
                "\n",
            )
        )
    return str_builder


def format_value_removed(deep_diff: dict[str, dict]) -> list[str]:
    """
    Formats a dict of values that were removed to a list of strings

    >>> deep_diff = {
    ...     "root[0]": {
    ...         "description": "None",
    ...         "name": "Name"
    ...     },
    ... }
    >>> format_value_removed(deep_diff)
    [' * [0]:\\n', "\\t- {'description': 'None', 'name': 'Name'}\\n"]

    """
    str_builder = []
    for _path, removed in deep_diff.items():
        str_builder.extend(
            (
                f" * {_path.replace('root', '')}:\n",
                f"\t- {pformat(removed)}\n",
            )
        )
    return str_builder


def _get_from_deep_diff_path(deep_diff_path: str, lookup_model: dict) -> Any:
    """
    Similar to `format_deep_diff_path` and `get_dict_dot_keys` in
    `cognite.powerops.clients.shop.data_classes.helpers` but modified
    to work with the deepdiff format separated from yaml formats

    >>> _path = "root['key_1'][0]['key_3']"
    >>> _lookup_model = {
    ...    "key_1": [{"key_2": "value_2", "key_3": "value_3"}]
    ... }
    >>> _get_from_deep_diff_path(_path, _lookup_model)
    'value_3'

    """
    keys = [
        int(k) if k.isdigit() else k
        for k in deep_diff_path.replace("root[", "").replace("'", "").removesuffix("]").split("][")
    ]
    try:
        item = reduce(operator.getitem, keys, lookup_model)
    except KeyError:
        item = f"Could not retrieve at {deep_diff_path}"
    return item


def format_value_added(deep_diff: PrettyOrderedSet, lookup_model: dict) -> list[str]:
    """
    Formats a dict of values that were added to a list of strings
    The deep_diff does not contain the new value, so it is fetched from the lookup_model
    """
    str_builder = []
    _path: str = None  # type: ignore
    for _path in deep_diff:
        str_builder.extend(
            (
                f" * {_path.replace('root', '')}:\n",
                f"\t- {pformat(_get_from_deep_diff_path(_path, lookup_model))}\n",
            )
        )
    return str_builder


if __name__ == "__main__":
    doctest.testmod()



"""
The module contains functions for transforming the ReSync configuration into Cognite Data Model Types which are used
in multiple models. Note the use of the `make_ext_id` function which is used to generate unique external IDs for
all the conversions. This is done to ensure that the same external ID is used for the same object across multiple
transformations.
"""

from __future__ import annotations

from cognite.powerops.clients.data_classes import (
    DateTransformationApply,
    InputTimeSeriesMappingApply,
    ScenarioMappingApply,
    ShopTransformationApply,
    ValueTransformationApply,
)
from cognite.powerops.resync.config.market._core import RelativeTime
from cognite.powerops.resync.config.shared import TimeSeriesMapping, TimeSeriesMappingEntry, Transformation
from cognite.powerops.resync.utils.common import make_ext_id


def _to_date_transformations(time: RelativeTime | list[DateTransformationApply]) -> list[DateTransformationApply]:
    if isinstance(time, list) and time and isinstance(time[0], DateTransformationApply):
        return time
    output = []
    for operation in time.operations or []:
        method, arguments = operation
        args, kwargs = [], {}
        if isinstance(arguments, dict):
            kwargs = arguments
        elif isinstance(arguments, list):
            args = arguments
        elif isinstance(arguments, str):
            args = [arguments]
        else:
            raise ValueError(f"Unknown arguments type: {type(arguments)}")
        output.append(
            DateTransformationApply(
                external_id=make_ext_id(operation, DateTransformationApply),
                transformation=method,
                args=args,
                kwargs=kwargs,
            )
        )
    return output


def _to_shop_transformation(
    start: RelativeTime | list[DateTransformationApply], end: RelativeTime | list[DateTransformationApply]
) -> ShopTransformationApply:
    start_list = _to_date_transformations(start)
    end_list = _to_date_transformations(end)
    external_id = make_ext_id([d.model_dump_json() for d in start_list + end_list], ShopTransformationApply)
    return ShopTransformationApply(
        external_id=external_id,
        type_name=ShopTransformationApply.__name__.removesuffix("Apply"),
        start=start_list,
        end=end_list,
    )


def _to_input_timeseries_mapping(entry: TimeSeriesMappingEntry) -> InputTimeSeriesMappingApply:
    return InputTimeSeriesMappingApply(
        external_id=make_ext_id(entry.model_dump_json(), class_=InputTimeSeriesMappingApply),
        aggregation=entry.aggregation.name if entry.aggregation else None,
        cdf_time_series=entry.time_series_external_id,
        shop_object_type=entry.object_type,
        shop_object_name=entry.object_name,
        shop_attribute_name=entry.attribute_name,
        retrieve=entry.retrieve.name if entry.retrieve else None,
        transformations=[_to_value_transformation(t) for t in (entry.transformations or [])],
    )


def _to_value_transformation(transformation: Transformation) -> ValueTransformationApply:
    return ValueTransformationApply(
        external_id=make_ext_id(transformation.model_dump_json(), class_=ValueTransformationApply),
        method=transformation.transformation.name,
        arguments=transformation.kwargs,
    )


def _to_scenario_mapping(external_id: str, name: str, time_series_mapping: TimeSeriesMapping) -> ScenarioMappingApply:
    mappings = [_to_input_timeseries_mapping(entry) for entry in time_series_mapping]

    return ScenarioMappingApply(
        external_id=external_id,
        mapping_override=mappings,
        name=name,
    )



from __future__ import annotations

import abc
from collections import defaultdict, Counter
from itertools import groupby
import json
from abc import ABC

from cognite.client.exceptions import CogniteNotFoundError
from deepdiff import DeepDiff
from deepdiff.model import PrettyOrderedSet

from pathlib import Path
from typing import Any, ClassVar, Iterable, Optional, Union, TypeVar, get_args
from typing import Type as TypingType
from pydantic import BaseModel, ConfigDict

from cognite.client import CogniteClient
from cognite.client.data_classes import Asset, Label, Relationship, TimeSeries
from cognite.client.data_classes.data_modeling.instances import (
    EdgeApply,
    NodeApply,
    InstancesApply,
    NodeApplyList,
    EdgeApplyList,
)

from cognite.powerops.cdf_labels import AssetLabel, RelationshipLabel
from cognite.powerops.clients.powerops_client import PowerOpsClient
from cognite.powerops.clients.data_classes._core import DomainModelApply
from cognite.powerops.cogshop1.data_classes._core import DomainModelApply as DomainModelApplyCogShop1
from cognite.powerops.resync.models.cdf_resources import CDFFile, CDFSequence
from cognite.powerops.resync.models.helpers import (
    format_change_binary,
    format_value_added,
    format_value_removed,
    isinstance_list,
    match_field_from_relationship,
    pydantic_model_class_candidate,
)
from cognite.powerops.resync.utils.common import all_subclasses
from cognite.powerops.resync.utils.serializer import remove_read_only_fields

_T_Type = TypeVar("_T_Type")


class ResourceType(BaseModel, ABC):
    def sequences(self) -> list[CDFSequence]:
        return self._fields_of_type(CDFSequence)

    def files(self) -> list[CDFFile]:
        return self._fields_of_type(CDFFile)

    def time_series(self) -> list[TimeSeries]:
        return self._fields_of_type(TimeSeries)

    def _fields_of_type(self, type_: TypingType[_T_Type]) -> list[_T_Type]:
        output: list[_T_Type] = []
        for field_name in self.model_fields:
            value = getattr(self, field_name)
            if not value:
                continue
            elif isinstance_list(value, type_):
                output.extend(value)
            elif isinstance(value, type_):
                output.append(value)
        return output


class AssetType(ResourceType, ABC):
    parent_external_id: ClassVar[str]
    label: ClassVar[Union[AssetLabel, str]]
    model_config: ClassVar[ConfigDict] = ConfigDict(arbitrary_types_allowed=True)
    parent_description: ClassVar[Optional[str]] = None
    _instantiated_assets: ClassVar[dict[str, AssetType]] = defaultdict(dict)

    name: str
    description: Optional[str] = None
    _external_id: Optional[str] = None
    _type: Optional[str] = None

    @property
    def external_id(self) -> str:
        return self._external_id or f"{self.type_}_{self.name}"

    @external_id.setter
    def external_id(self, value: str) -> None:
        """
        This setter is only used in the Market models which have inconsistent naming between the type and the
        parent_external_id. It is a workaround to keep the refactoring to models to avoid introducing breaking changes
        in CDF.
        Parameters
        ----------
        value : str
            The external id of the asset.

        """
        self._external_id = value

    @property
    def type_(self) -> str:
        return self._type or self.parent_external_id.removesuffix("s")

    @type_.setter
    def type_(self, value: str) -> None:
        """
        This setter is only used in the Market models which have inconsistent naming between the type and
        parent_external_id. It is a workaround to keep the refactoring to models to avoid introducing breaking changes
        in CDF.

        Parameters
        ----------
        value: str
            The type of the asset.

        """
        self._type = value

    @property
    def parent_name(self):
        return self.parent_external_id.replace("_", " ").title()

    def relationships(self) -> list[Relationship]:
        return self._relationships(self)

    def _relationships(self, to: BaseModel) -> list[Relationship]:
        relationships = []
        for field_name, field in to.model_fields.items():
            field_value = getattr(to, field_name)
            if not field_value:
                continue
            if isinstance(field_value, CDFSequence):
                relationships.append(self._create_relationship(field_value.external_id, "SEQUENCE", field_name))
            elif isinstance(field_value, TimeSeries):
                relationships.append(self._create_relationship(field_value.external_id, "TIMESERIES", field_name))
            elif isinstance(field_value, list) and field_value and isinstance(field_value[0], TimeSeries):
                # The string comparison is to avoid circular imports
                if type(to).__name__ == "ProductionPlanTimeSeries" and field_name == "series":
                    # For some reason, the production plan time series are not linked to the benchmarking config
                    continue

                relationships.extend(
                    self._create_relationship(ts.external_id, "TIMESERIES", field_name) for ts in field_value
                )
            elif isinstance(field_value, list) and field_value and isinstance(field_value[0], CDFSequence):
                relationships.extend(
                    self._create_relationship(sequence.external_id, "SEQUENCE", field_name) for sequence in field_value
                )
            elif isinstance(field_value, AssetType) or (
                isinstance(field_value, list) and field_value and isinstance(field_value[0], AssetType)
            ):
                asset_types: list[AssetType] = field_value if isinstance(field_value, list) else [field_value]
                for target in asset_types:
                    target_type = target.type_
                    if self.type_ == "plant" and target.type_ == "reservoir":
                        target_type = "inlet_reservoir"
                    relationships.append(self._create_relationship(target.external_id, "ASSET", target_type))
            elif isinstance(field_value, NonAssetType) or (
                isinstance(field_value, list) and field_value and isinstance(field_value[0], NonAssetType)
            ):
                non_asset_types: list[NonAssetType] = field_value if isinstance(field_value, list) else [field_value]
                for target in non_asset_types:
                    relationships.extend(self._relationships(target))

        return relationships

    def _create_relationship(
        self,
        target_external_id: str,
        target_cdf_type: str,
        target_type: str,
    ) -> Relationship:
        source_type = "ASSET"

        # The market model uses the suffix CDF type for the relationship label, while the core model does not.
        try:
            # Production Model
            label = RelationshipLabel(f"relationship_to.{target_type}")
        except ValueError:
            # Market Model
            label = RelationshipLabel(f"relationship_to.{target_type}_{target_cdf_type.lower()}")
            # In addition, the Market Model uses capitalised CDF types for the relationship type,
            # while the core model uses all upper cases.
            source_type = source_type.title()
            target_cdf_type = target_cdf_type.title()

        return Relationship(
            external_id=f"{self.external_id}.{target_external_id}",
            source_external_id=self.external_id,
            source_type=source_type,
            target_external_id=target_external_id,
            target_type=target_cdf_type,
            labels=[Label(external_id=label.value)],
        )

    def as_asset(self):
        metadata = {}
        for field_name, field in self.model_fields.items():
            if (
                any(cdf_type in str(field.annotation) for cdf_type in [CDFSequence.__name__, TimeSeries.__name__])
                or field_name in {"name", "description", "label", "parent_external_id"}
                or field.exclude
            ):
                continue
            value = getattr(self, field_name)
            if value is None or isinstance(value, AssetType) or isinstance_list(value, AssetType):
                continue
            if isinstance(value, list) and not value:
                continue
            if isinstance(value, NonAssetType):
                value = value.model_dump(exclude_unset=True)
                for k, v in value.items():
                    if isinstance(v, (dict, list)):
                        v = json.dumps(v)
                    metadata[f"{field_name}:{k}"] = v
                continue

            if isinstance(value, (dict, list)):
                value = json.dumps(value)
            metadata[field_name] = value

        return Asset(
            external_id=self.external_id,
            name=self.name,
            parent_external_id=self.parent_external_id,
            labels=[Label(self.label.value)],
            metadata=metadata or None,
            description=self.description,
        )

    @classmethod
    def _parse_asset_metadata(cls, metadata: dict[str, str] | None = None) -> dict[str, dict | str]:
        if not metadata:
            return {}
        output: dict[str, dict | str] = {}
        for key, value in metadata.items():
            counts = Counter(key)
            if counts[":"] == 1:
                field_name, sub_key = key.split(":")
                if field_name not in output:
                    output[field_name] = {}
                output[field_name][sub_key] = value
            elif counts[":"] == 0:
                output[key] = value
            else:
                raise NotImplementedError("Nested keys more than one level is not supported")

        return output

    @classmethod
    def _from_asset(
        cls,
        asset: Asset,
        additional_fields: Optional[dict[str, Any]] = None,
    ) -> T_Asset_Type:
        if not additional_fields:
            additional_fields = {}
        metadata = cls._parse_asset_metadata(asset.metadata)
        instance = cls(
            _external_id=asset.external_id,
            name=asset.name,
            description=asset.description,
            **metadata,
            **additional_fields,
        )
        AssetType._instantiated_assets[asset.external_id] = instance
        return instance

    @classmethod
    def _field_name_asset_resource_class(cls) -> Iterable[tuple[str, TypingType[AssetType]]]:
        """AssetType fields are are of type list[AssetType] or Optional[AssetType]"""
        # todo? method is identical to _field_name_asset_resource_class in AssetModel
        # * unsure how to reuse?
        for field_name in cls.model_fields:
            class_ = cls.model_fields[field_name].annotation
            if pydantic_model_class_candidate(class_):
                asset_resource_class = get_args(class_)[0]
                if issubclass(asset_resource_class, AssetType):
                    yield field_name, asset_resource_class

    @classmethod
    def _fetch_and_set_metadata(
        cls,
        client: CogniteClient,
        data_set_id: int,
        additional_fields: dict[str, Union[list, None]],
        asset_external_id: str,
        fetch_metadata: bool,
        fetch_content: bool,
    ) -> dict[str, Any]:
        """Fetches resources that are linked with relationships to the asset."""
        relationships = client.relationships.list(
            source_external_ids=[asset_external_id],
            source_types=["asset"],
            target_types=["timeseries", "asset", "sequence", "file"],
            data_set_ids=[data_set_id],
            limit=-1,
        )
        for r in relationships:
            field = match_field_from_relationship(cls.model_fields, r)
            target_type = r.target_type.lower()

            if target_type == "asset":
                if r.target_external_id in AssetType._instantiated_assets:
                    relationship_target = AssetType._instantiated_assets[r.target_external_id]

                else:
                    target_class = [y for x, y in cls._field_name_asset_resource_class() if field == x and y][0]
                    relationship_target = target_class.from_cdf(
                        client=client,
                        data_set_id=data_set_id,
                        external_id=r.target_external_id,
                        fetch_metadata=fetch_metadata,
                        fetch_content=fetch_content,
                    )

            elif target_type == "timeseries":
                relationship_target = client.time_series.retrieve(external_id=r.target_external_id)
            elif target_type == "sequence":
                try:
                    relationship_target = CDFSequence.from_cdf(client, r.target_external_id, fetch_content)
                except CogniteNotFoundError:
                    relationship_target = None
            elif target_type == "file":
                try:
                    relationship_target = CDFFile.from_cdf(client, r.target_external_id, fetch_content)
                except CogniteNotFoundError:
                    relationship_target = None
            else:
                raise ValueError(f"Cannot handle target type {r.target_type}")

            # Add relationship target to additional fields in-place
            if isinstance(additional_fields[field], list):
                additional_fields[field].append(relationship_target)
            else:
                additional_fields[field] = relationship_target

    @classmethod
    def from_cdf(
        cls,
        client: CogniteClient,
        data_set_id: int,
        external_id: Optional[str] = "",
        asset: Optional[Asset] = None,
        fetch_metadata: bool = True,
        fetch_content: bool = False,
    ) -> T_Asset_Type:
        """
        Fetch an asset from CDF and convert it to a model instance.
        Optionally fetch relationships targets and content by setting
        `fetch_metadata` and optionally `fetch_content`

        By default, content of files/sequences/time series is not fetched.
        This can be enabled by setting `fetch_content=True`.
        """

        if asset and external_id:
            raise ValueError("Only one of asset and external_id can be provided")
        if external_id:
            # Check if asset has already been instantiated, eg. by a relationship
            if external_id in AssetType._instantiated_assets:
                return AssetType._instantiated_assets[external_id]
            else:
                asset = client.assets.retrieve(
                    external_id=external_id,
                )
        if not asset:
            raise ValueError(f"Could not retrieve asset with {external_id=}")

        # Prepare non-asset metadata fields
        additional_fields = {
            field: [] if "list" in str(field_info.annotation) else None
            for field, field_info in cls.model_fields.items()
            if field in [x for x, _ in cls._field_name_asset_resource_class()]
            or any(cdf_type in str(field_info.annotation) for cdf_type in [CDFSequence.__name__, TimeSeries.__name__])
        }

        # Populate non-asset metadata fields, according to relationships/flags
        # `Additional_fields`, is modified in-place by `_fetch_metadata`
        if fetch_metadata:
            cls._fetch_and_set_metadata(
                client,
                data_set_id,
                additional_fields,
                asset.external_id,
                fetch_metadata,
                fetch_content,
            )
        instance = cls._from_asset(asset, additional_fields)
        if asset.external_id != instance.external_id:
            # The external id is not set in a standardized way for Market configs.
            instance.external_id = asset.external_id
        return instance

    def _asset_type_prepare_for_diff(self: T_Asset_Type) -> dict[str, dict]:
        for model_field in self.model_fields:
            field = getattr(self, model_field)
            if isinstance(field, AssetType):
                # Only include external id in diff
                setattr(self, model_field, field.external_id)

            elif isinstance_list(field, AssetType):
                # Sort bt external id to have consistent order for diff
                setattr(self, model_field, sorted(map(lambda x: x.external_id, field)))
            elif isinstance(field, Path):
                # remove path from diff
                setattr(self, model_field, None)
        return self


T_Asset_Type = TypeVar("T_Asset_Type", bound=AssetType)


class NonAssetType(BaseModel, ABC):
    model_config: ClassVar[ConfigDict] = ConfigDict(arbitrary_types_allowed=True)


class Model(BaseModel, ABC):
    def sequences(self) -> list[CDFSequence]:
        sequences = [sequence for item in self._resource_types() for sequence in item.sequences()]
        sequences.extend(self._fields_of_type(CDFSequence))
        return sequences

    def files(self) -> list[CDFFile]:
        files = [file for item in self._resource_types() for file in item.files()]
        files.extend(self._fields_of_type(CDFFile))
        return files

    def time_series(self) -> list[TimeSeries]:
        time_series = [ts for item in self._resource_types() for ts in item.time_series()]
        time_series.extend(self._fields_of_type(TimeSeries))
        return time_series

    def drop_content(self) -> None:
        for item in self._resource_types():
            for sequence in item.sequences():
                sequence.content = None
            for file in item.files():
                file.content = None

    def _resource_types(self) -> Iterable[ResourceType]:
        for f in self.model_fields:
            if isinstance(items := getattr(self, f), list) and items and isinstance(items[0], ResourceType):
                yield from items

    def _fields_of_type(self, type_: TypingType[_T_Type]) -> Iterable[_T_Type]:
        for field_name in self.model_fields:
            value = getattr(self, field_name)
            if isinstance(value, type_):
                yield value
            elif isinstance_list(value, type_):
                yield from value

    @property
    def model_name(self) -> str:
        return type(self).__name__

    def summary(self) -> dict[str, dict[str, dict[str, int]]]:
        summary: dict[str, dict[str, dict[str, int]]] = {self.model_name: {"domain": {}, "cdf": {}}}
        summary[self.model_name]["domain"] = {
            field_name: len(value) if isinstance(value := getattr(self, field_name), (list, dict)) else 1
            for field_name in self.model_fields
        }
        summary[self.model_name]["cdf"]["files"] = len(self.files())
        summary[self.model_name]["cdf"]["sequences"] = len(self.sequences())
        summary[self.model_name]["cdf"]["time_series"] = len(self.time_series())
        return summary

    def dump_external_ids(self) -> dict[str, dict, str, dict[str, list[str]]]:
        output: dict[str, dict, str, dict[str, list[str]]] = {self.model_name: {"domain": {}, "cdf": {}}}
        output[self.model_name]["domain"] = {
            field_name: [(v.external_id if hasattr(v, "external_id") else str(v)) for v in value]
            if isinstance(value := getattr(self, field_name), (list, dict))
            else [value.external_id if hasattr(value, "external_id") else str(value)]
            for field_name in self.model_fields
        }
        output[self.model_name]["cdf"]["files"] = [file.external_id for file in self.files()]
        output[self.model_name]["cdf"]["sequences"] = [sequence.external_id for sequence in self.sequences()]
        output[self.model_name]["cdf"]["time_series"] = [time_series.external_id for time_series in self.time_series()]
        return output

    def summary_diff(self: T_Model, other: T_Model) -> dict[str, dict[str, dict[str, dict[str, int]]]]:
        this_summary = self.dump_external_ids()
        other_summary = other.dump_external_ids()
        output: dict[str, dict[str, dict[str, dict[str, int]]]] = {}
        for model_name in this_summary:
            output[model_name] = {}
            for domain in this_summary[model_name]:
                output[model_name][domain] = {}
                for type_ in this_summary[model_name][domain]:
                    this_ext_ids = set(this_summary[model_name][domain][type_])
                    other_ext_ids = set(other_summary[model_name][domain][type_])
                    if not type_ == "nodes":
                        output[model_name][domain][type_] = {
                            "added": len(this_ext_ids - other_ext_ids),
                            "removed": len(other_ext_ids - this_ext_ids),
                        }
                        continue
                    # Todo this is a hack go get more detailed diff for nodes
                    # To make it not a hack the external id has to be standardized
                    # and not just assumed to have this format.
                    output[model_name][domain][type_] = {}

                    this_ext_ids_by_node_type = defaultdict(set)
                    for node_type, ext_id in groupby(sorted(this_ext_ids), key=lambda x: x.split("_", 1)[0]):
                        this_ext_ids_by_node_type[node_type].update(ext_id)
                    other_ext_ids_by_node_type = defaultdict(set)
                    for node_type, ext_id in groupby(sorted(other_ext_ids), key=lambda x: x.split("_", 1)[0]):
                        other_ext_ids_by_node_type[node_type].update(ext_id)
                    for node_type in set(this_ext_ids_by_node_type) | set(other_ext_ids_by_node_type):
                        if node_type != "BM":
                            output[model_name][domain][type_][node_type] = {
                                "added": len(
                                    this_ext_ids_by_node_type[node_type] - other_ext_ids_by_node_type[node_type]
                                ),
                                "removed": len(
                                    other_ext_ids_by_node_type[node_type] - this_ext_ids_by_node_type[node_type]
                                ),
                            }
                            continue
                        output[model_name][domain][type_][node_type] = {}
                        this_mapping_by_watercourse = defaultdict(set)
                        for watercourse, ext_id in groupby(
                            sorted(this_ext_ids_by_node_type[node_type]), lambda x: x.split("__")[1]
                        ):
                            this_mapping_by_watercourse[watercourse].update(ext_id)
                        other_mapping_by_watercourse = defaultdict(set)
                        for watercourse, ext_id in groupby(
                            sorted(other_ext_ids_by_node_type[node_type]), lambda x: x.split("__")[1]
                        ):
                            other_mapping_by_watercourse[watercourse].update(ext_id)
                        for watercourse in set(this_mapping_by_watercourse) | set(other_mapping_by_watercourse):
                            output[model_name][domain][type_][node_type][watercourse] = {
                                "added": len(
                                    this_mapping_by_watercourse[watercourse] - other_mapping_by_watercourse[watercourse]
                                ),
                                "removed": len(
                                    other_mapping_by_watercourse[watercourse] - this_mapping_by_watercourse[watercourse]
                                ),
                            }
        return output

    def dump(self) -> dict[str, Any]:
        output: dict[str, Any] = {}
        if time_series := self.time_series():
            output["time_series"] = sorted(
                ({"external_id": ts.external_id} if isinstance(ts, TimeSeries) else ts for ts in time_series),
                key=self._external_id_key,
            )
        if files := self.files():
            output["files"] = sorted(
                (remove_read_only_fields(file.dump(camel_case=False)) for file in files),
                key=self._external_id_key,
            )
        if sequences := self.sequences():

            def dump_sequence(resource: CDFSequence) -> dict[str, Any]:
                output = remove_read_only_fields(resource.dump(camel_case=False))
                if (columns := output.get("columns")) and isinstance(columns, list):
                    for no, column in enumerate(columns):
                        output["columns"][no] = remove_read_only_fields(column)
                return output

            output["sequences"] = sorted(
                (dump_sequence(sequence) for sequence in sequences),
                key=self._external_id_key,
            )
        return output

    @classmethod
    def load(cls: TypingType[T_Model], data: dict[str, Any]) -> T_Model:
        raise NotImplementedError()

    @classmethod
    def _external_id_key(cls, resource: dict | Any) -> str:
        if hasattr(resource, "external_id"):
            return resource.external_id.casefold()
        elif isinstance(resource, dict) and ("external_id" in resource or "externalId" in resource):
            return resource.get("external_id", resource.get("externalId")).casefold()
        raise ValueError(f"Could not find external_id in {resource}")

    @classmethod
    @abc.abstractmethod
    def from_cdf(
        cls: TypingType[T_Model],
        client: PowerOpsClient,
        data_set_id: int,
        fetch_metadata: bool = True,
        fetch_content: bool = False,
    ) -> T_Model:
        ...

    def difference(self: T_Model, other: T_Model, print_string: bool = True) -> dict:
        if type(self) != type(other):
            raise ValueError("Cannot compare these models of different types.")

        self_dump = self._prepare_for_diff()
        other_dump = other._prepare_for_diff()
        diff_dict = {}
        for model_field in self_dump:
            if deep_diff := DeepDiff(
                self_dump[model_field],
                other_dump[model_field],
                ignore_type_in_groups=[(float, int, type(None))],
                exclude_regex_paths=[
                    r"(.+?)._cognite_client",
                    r"(.+?).last_updated_time",
                    r"(.+?).parent_id",
                    r"(.+?).root_id]",
                    r"(.+?).data_set_id",
                    r"(.+?).created_time",
                    r"(.+?)lastUpdatedTime",
                    r"(.+?)createdTime",
                    r"(.+?)parentId",
                    r"(.+?)\.id",
                    # Relevant metadata should already be included in the model
                    r"(.+?)metadata",
                ],
            ).to_dict():
                diff_dict[model_field] = deep_diff

        if print_string:
            _diff_formatter = _DiffFormatter(
                full_diff_per_field=diff_dict,
                model_a=self_dump,
                model_b=other_dump,
            )
            print(_diff_formatter.format_as_string())

        return diff_dict

    def difference_external_ids(self: T_Model, other: T_Model) -> dict:
        this_summary = self.dump_external_ids()
        other_summary = other.dump_external_ids()
        output: dict[str, dict[str, dict[str, dict[str, list[str]]]]] = {}
        for model_name in this_summary:
            output[model_name] = {}
            for domain in this_summary[model_name]:
                output[model_name][domain] = {}
                for type_ in this_summary[model_name][domain]:
                    this_ext_ids = set(this_summary[model_name][domain][type_])
                    other_ext_ids = set(other_summary[model_name][domain][type_])
                    if not type_ == "nodes":
                        added = sorted(this_ext_ids - other_ext_ids)
                        removed = sorted(other_ext_ids - this_ext_ids)
                        output[model_name][domain][type_] = {
                            "added": added[: min(len(added), 10)],
                            "removed": removed[: min(len(removed), 10)],
                        }
                        continue
                    # Todo this is a hack go get more detailed diff for nodes
                    # To make it not a hack the external id has to be standardized
                    # and not just assumed to have this format.
                    output[model_name][domain][type_] = {}

                    this_ext_ids_by_node_type = defaultdict(set)
                    for node_type, ext_id in groupby(sorted(this_ext_ids), key=lambda x: x.split("_", 1)[0]):
                        this_ext_ids_by_node_type[node_type].update(ext_id)
                    other_ext_ids_by_node_type = defaultdict(set)
                    for node_type, ext_id in groupby(sorted(other_ext_ids), key=lambda x: x.split("_", 1)[0]):
                        other_ext_ids_by_node_type[node_type].update(ext_id)
                    for node_type in set(this_ext_ids_by_node_type) | set(other_ext_ids_by_node_type):
                        if node_type != "BM":
                            added = sorted(this_ext_ids_by_node_type[node_type] - other_ext_ids_by_node_type[node_type])
                            removed = sorted(
                                other_ext_ids_by_node_type[node_type] - this_ext_ids_by_node_type[node_type]
                            )
                            output[model_name][domain][type_][node_type] = {
                                "added": added[: min(len(added), 10)],
                                "removed": removed[: min(len(removed), 10)],
                            }
                            continue
                        output[model_name][domain][type_][node_type] = {}
                        this_mapping_by_watercourse = defaultdict(set)
                        for watercourse, ext_id in groupby(
                            sorted(this_ext_ids_by_node_type[node_type]), lambda x: x.split("__")[1]
                        ):
                            this_mapping_by_watercourse[watercourse].update(ext_id)
                        other_mapping_by_watercourse = defaultdict(set)
                        for watercourse, ext_id in groupby(
                            sorted(other_ext_ids_by_node_type[node_type]), lambda x: x.split("__")[1]
                        ):
                            other_mapping_by_watercourse[watercourse].update(ext_id)
                        for watercourse in set(this_mapping_by_watercourse) | set(other_mapping_by_watercourse):
                            added = sorted(
                                this_mapping_by_watercourse[watercourse] - other_mapping_by_watercourse[watercourse]
                            )
                            removed = sorted(
                                other_mapping_by_watercourse[watercourse] - this_mapping_by_watercourse[watercourse]
                            )

                            output[model_name][domain][type_][node_type][watercourse] = {
                                "added": added[: min(len(added), 10)],
                                "removed": removed[: min(len(removed), 10)],
                            }
        return output


T_Model = TypeVar("T_Model", bound=Model)


class AssetModel(Model, ABC):
    root_asset: ClassVar[Optional[Asset]] = None

    def assets(self) -> list[Asset]:
        return [item.as_asset() for item in self._asset_types()]

    def relationships(self) -> list[Relationship]:
        return [edge for item in self._asset_types() for edge in item.relationships()]

    def parent_assets(self) -> list[Asset]:
        if not self.root_asset:
            return []

        def _to_name(external_id: str) -> str:
            parts = external_id.replace("_", " ").split(" ")
            if parts[0].lower() == "rkom":
                parts[0] = parts[0].upper()
            else:
                parts[0] = parts[0].title()
            # The replacing is to have avoid doing changes, even though this is inconsistent with the other parent
            # assets.
            return " ".join(parts).replace("Bid process", "Bid").replace("bid process", "bid")

        parent_and_description_ids = {
            (item.parent_external_id, item.parent_description or "") for item in self._asset_types()
        }

        return [self.root_asset] + [
            Asset(
                external_id=parent_id,
                name=_to_name(parent_id),
                parent_external_id=self.root_asset.external_id,
                description=description,
            )
            for parent_id, description in parent_and_description_ids
        ]

    def _asset_types(self) -> Iterable[AssetType]:
        yield from (item for item in self._resource_types() if isinstance(item, AssetType))

    @classmethod
    def _field_name_asset_resource_class(cls) -> Iterable[tuple[str, TypingType[AssetType]]]:
        """AssetType fields are of type list[AssetType] or Optional[AssetType]"""
        # todo? method is identical to _field_name_asset_resource_class in AssetType
        # * unsure how to reuse?
        for field_name in cls.model_fields:
            class_ = cls.model_fields[field_name].annotation
            if pydantic_model_class_candidate(class_):
                asset_resource_class = get_args(class_)[0]
                if (is_asset_type := issubclass(asset_resource_class, AssetType)) and any(
                    base is abc.ABC for base in asset_resource_class.__bases__
                ):
                    for subclass in all_subclasses(asset_resource_class):
                        yield field_name, subclass
                elif is_asset_type:
                    yield field_name, asset_resource_class

    def summary(self) -> dict[str, dict[str, dict[str, int]]]:
        summary = super().summary()
        summary[self.model_name]["cdf"]["assets"] = len(self.assets())
        summary[self.model_name]["cdf"]["relationships"] = len(self.relationships())
        summary[self.model_name]["cdf"]["parent_assets"] = len(self.parent_assets())
        return summary

    def dump_external_ids(self) -> dict[str, dict, str, dict[str, list[str]]]:
        output = super().dump_external_ids()
        output[self.model_name]["cdf"]["assets"] = [asset.external_id for asset in self.assets()]
        output[self.model_name]["cdf"]["relationships"] = [
            relationship.external_id for relationship in self.relationships()
        ]
        output[self.model_name]["cdf"]["parent_assets"] = [asset.external_id for asset in self.parent_assets()]
        return output

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        for field_method in [self.assets, self.relationships, self.parent_assets]:
            if resources := field_method():
                output[field_method.__name__] = sorted(
                    (remove_read_only_fields(resource.dump(camel_case=False)) for resource in resources),
                    key=self._external_id_key,
                )
        return output

    @classmethod
    def from_cdf(
        cls: TypingType[T_Asset_Model],
        client: PowerOpsClient,
        data_set_id: int,
        fetch_metadata: bool = True,
        fetch_content: bool = False,
    ) -> T_Asset_Model:
        if fetch_content and not fetch_metadata:
            raise ValueError("Cannot fetch content without also fetching metadata")
        cdf = client.cdf
        output = defaultdict(list)
        for field_name, asset_cls in cls._field_name_asset_resource_class():
            try:
                assets = cdf.assets.retrieve_subtree(external_id=asset_cls.parent_external_id)
            except AttributeError:
                raise
            for asset in assets:
                if asset.external_id == asset_cls.parent_external_id:
                    continue
                instance = asset_cls.from_cdf(
                    client=cdf,
                    data_set_id=data_set_id,
                    asset=asset,
                    fetch_metadata=fetch_metadata,
                    fetch_content=fetch_content,
                )
                output[field_name].append(instance)

        return cls(**output)

    def _prepare_for_diff(self: T_Asset_Model) -> dict[str:dict]:
        clone = self.model_copy(deep=True)

        for model_field in clone.model_fields:
            field_value = getattr(clone, model_field)
            if isinstance_list(field_value, AssetType):
                # Sort the asset types to have comparable order for diff
                _sorted = sorted(field_value, key=lambda x: x.external_id)
                # Prepare each asset type for diff
                _prepared = map(lambda x: x._asset_type_prepare_for_diff(), _sorted)
                setattr(clone, model_field, list(_prepared))
            elif isinstance(field_value, AssetType):
                field_value._asset_type_prepare_for_diff()
        # Some fields are have been set to their external_id which gives a warning we can ignore
        return clone.model_dump(warnings=False)

    def difference(self: T_Asset_Model, other: T_Asset_Model, print_string: bool = True) -> dict:
        if type(self) != type(other):
            raise ValueError("Cannot compare these models of different types.")

        self_dump = self._prepare_for_diff()
        other_dump = other._prepare_for_diff()
        diff_dict = {}
        for model_field in self_dump:
            if deep_diff := DeepDiff(
                self_dump[model_field],
                other_dump[model_field],
                ignore_type_in_groups=[(float, int, type(None))],
                exclude_regex_paths=[
                    r"(.+?)._cognite_client",
                    r"(.+?).last_updated_time",
                    r"(.+?).parent_id",
                    r"(.+?).root_id]",
                    r"(.+?).data_set_id",
                    r"(.+?).created_time",
                    r"(.+?)lastUpdatedTime",
                    r"(.+?)createdTime",
                    r"(.+?)parentId",
                    r"(.+?)\.id",
                    # Relevant metadata should already be included in the model
                    r"(.+?)metadata",
                ],
            ).to_dict():
                diff_dict[model_field] = deep_diff

        if print_string:
            _diff_formatter = _DiffFormatter(
                full_diff_per_field=diff_dict,
                model_a=self_dump,
                model_b=other_dump,
            )
            print(_diff_formatter.format_as_string())

        return diff_dict


T_Asset_Model = TypeVar("T_Asset_Model", bound=AssetModel)


class DataModel(Model, ABC):
    def instances(self) -> InstancesApply:
        nodes: dict[str, NodeApply] = {}
        edges: dict[str, EdgeApply] = {}
        for domain_model in self._domain_models():
            instance_applies = domain_model.to_instances_apply()
            # Caching in case recursive relationships are used.
            for node in instance_applies.nodes:
                if node.external_id not in nodes:
                    nodes[node.external_id] = node
            for edge in instance_applies.edges:
                if edge.external_id not in edges:
                    edges[edge.external_id] = edge

        return InstancesApply(nodes=NodeApplyList(nodes.values()), edges=EdgeApplyList(edges.values()))

    def _domain_models(self) -> Iterable[DomainModelApply]:
        for field_name in self.model_fields:
            items = getattr(self, field_name)
            if isinstance(items, list) and items and isinstance(items[0], (DomainModelApply, DomainModelApplyCogShop1)):
                yield from items
            if isinstance(items, (DomainModelApply, DomainModelApplyCogShop1)):
                yield items
            if (
                isinstance(items, dict)
                and items
                and isinstance(next(iter(items.values())), (DomainModelApply, DomainModelApplyCogShop1))
            ):
                yield from items.values()

    def summary(self) -> dict[str, dict[str, dict[str, int]]]:
        summary = super().summary()
        instances = self.instances()
        summary[self.model_name]["cdf"]["nodes"] = len(instances.nodes)
        summary[self.model_name]["cdf"]["edges"] = len(instances.edges)
        return summary

    def dump_external_ids(self) -> dict[str, dict, str, dict[str, list[str]]]:
        output = super().dump_external_ids()
        instances = self.instances()
        output[self.model_name]["cdf"]["nodes"] = [node.external_id for node in instances.nodes]
        output[self.model_name]["cdf"]["edges"] = [edge.external_id for edge in instances.edges]
        return output

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        instances = self.instances()
        for instance_type in ["nodes", "edges"]:
            if resources := getattr(instances, instance_type):
                output[instance_type] = sorted(
                    (remove_read_only_fields(resource.dump(camel_case=False)) for resource in resources),
                    key=self._external_id_key,
                )
        return output


class _DiffFormatter:
    def __init__(self, full_diff_per_field: dict[str, dict], model_a: dict, model_b: dict):
        self.full_diff_per_field = full_diff_per_field
        self.model_a = model_a
        self.model_b = model_b

        self.str_builder: list = None

    def _format_per_field(self, field_name: str, field_diff: dict[str, Union[dict, PrettyOrderedSet]]):
        self.str_builder.extend(
            (
                "\n\n========================== ",
                *field_name.title().split("_"),
                " ==========================\n",
            )
        )
        # Might need a better fallback for names
        self.str_builder.append("Indexes and names:\n\t")
        names = [
            f'{i}:{d.get("display_name", False) or d.get("name", "")}, ' for i, d in enumerate(self.model_a[field_name])
        ]

        self.str_builder.extend(names)
        self.str_builder.append("\n\n")

        for diff_type, diffs in field_diff.items():
            is_iterable = "iterable" in diff_type

            if diff_type in ("type_changes", "values_changed"):
                self.str_builder.extend(
                    (
                        f'The following values have changed {"type and value" if "type" in diff_type else ""}:\n',
                        *format_change_binary(diffs),
                        "\n",
                    ),
                )
            elif "removed" in diff_type:
                self.str_builder.extend(
                    (
                        f"The following {'values' if is_iterable else 'entries'} have been removed:\n",
                        *format_value_removed(diffs),
                        "\n",
                    )
                )
            elif "added" in diff_type:
                self.str_builder.extend(
                    (
                        f"The following {'values' if is_iterable else 'entries'} have been added:\n",
                        *format_value_added(diffs, self.model_b[field_name]),
                        "\n",
                    )
                )

            else:
                print(f"cannot handle {diff_type=}")

    def format_as_string(self) -> str:
        self.str_builder = []
        for field_name, field_diff in self.full_diff_per_field.items():
            self._format_per_field(field_name, field_diff)

        return "".join(self.str_builder)



from __future__ import annotations

from typing import ClassVar, Optional

from cognite.client import data_modeling as dm
from pydantic import Field

from cognite.powerops.clients.data_classes._core import DomainModel, DomainModelApply, TypeList

__all__ = ["RKOMCombinationBid", "RKOMCombinationBidApply", "RKOMCombinationBidList"]


class RKOMCombinationBid(DomainModel):
    space: ClassVar[str] = "power-ops"
    auction: Optional[str] = None
    name: Optional[str] = None
    rkom_bid_configs: list[str] = Field([], alias="rkomBidConfigs")


class RKOMCombinationBidApply(DomainModelApply):
    space: ClassVar[str] = "power-ops"
    auction: Optional[str] = None
    name: Optional[str] = None
    rkom_bid_configs: list[str] = []

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        source = dm.NodeOrEdgeData(
            source=dm.ContainerId("power-ops", "RKOMCombinationBid"),
            properties={
                "auction": self.auction,
                "name": self.name,
                "rkomBidConfigs": self.rkom_bid_configs,
            },
        )
        sources.append(source)

        this_node = dm.NodeApply(
            space=self.space,
            external_id=self.external_id,
            existing_version=self.existing_version,
            sources=sources,
        )
        nodes = [this_node]
        edges = []
        cache.add(self.external_id)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))


class RKOMCombinationBidList(TypeList[RKOMCombinationBid]):
    _NODE = RKOMCombinationBid



from __future__ import annotations

from datetime import datetime, timezone
from pathlib import Path
from typing import Optional, Callable, overload, Any, Type
from uuid import uuid4

from cognite.client import CogniteClient
from cognite.client.data_classes import Event

from cognite.powerops.clients.powerops_client import get_powerops_client, PowerOpsClient
from cognite.powerops.resync._logger import configure_debug_logging
from cognite.powerops.resync.config.resource_collection import ResourceCollection
from cognite.powerops.resync.config.resync_config import ReSyncConfig
from cognite.powerops.resync.models.base import Model, AssetModel
from cognite.powerops.resync import models
from cognite.powerops.resync.to_models.transform import transform
from yaml import safe_dump
from cognite.powerops.resync.utils.common import all_concrete_subclasses
from cognite.powerops.utils.cdf import Settings

MODEL_BY_NAME: dict[str, Type[Model]] = {
    model.__name__: model for model in all_concrete_subclasses(Model)  # type: ignore[type-abstract]
}
AVAILABLE_MODELS: frozenset[str] = frozenset(MODEL_BY_NAME)
DEFAULT_MODELS: frozenset[str] = frozenset(
    [m.__name__ for m in [models.ProductionModel, models.MarketModel, models.CogShop1Asset]]
)


def plan(
    path: Path,
    market: str,
    echo: Optional[Callable[[str], None]] = None,
    model_names: Optional[str | list[str]] = None,
    dump_folder: Optional[Path] = None,
    echo_pretty: Optional[Callable[[Any], None]] = None,
    client: PowerOpsClient | None = None,
) -> None:
    echo = echo or print
    echo_pretty: Callable[[Any], None] = echo_pretty or echo
    settings = Settings()
    client = client or get_powerops_client()

    bootstrap_resources, config, models = _load_transform(market, path, client.cdf.config.project, echo, model_names)
    _remove_non_existing_relationship_time_series_targets(client.cdf, models, bootstrap_resources, echo)
    echo(f"Load transform completed, models {', '.join([type(m).__name__ for m in models])} loaded")

    data_set = client.cdf.data_sets.retrieve(external_id=settings.powerops.read_dataset)
    if not data_set or not data_set.id:
        raise ValueError(f"Data set with external_id {settings.powerops.read_dataset} not found in CDF")
    data_set_id = data_set.id
    for model in models:
        echo(f"Retrieving {type(model).__name__} from CDF")
        cdf_model = type(model).from_cdf(client, fetch_metadata=True, fetch_content=False, data_set_id=data_set_id)

        summary_diff = model.summary_diff(cdf_model)
        echo(f"Summary diff for {model.model_name}")
        echo_pretty(summary_diff)

        if dump_folder:
            dump_folder.mkdir(parents=True, exist_ok=True)

            (dump_folder / f"{model.model_name}_local.yaml").write_text(safe_dump(model.dump()))
            (dump_folder / f"{model.model_name}_cdf.yaml").write_text(safe_dump(cdf_model.dump()))
        external_ids_diff = model.difference_external_ids(cdf_model)
        echo(f"External ids diff for {model.model_name}")
        echo_pretty(external_ids_diff)


@overload
def apply(
    path: Path,
    market: str,
    model_names: str,
    echo: Optional[Callable[[Any], None]] = None,
    auto_yes: bool = False,
    echo_pretty: Optional[Callable[[Any], None]] = None,
) -> Model:
    ...


@overload
def apply(
    path: Path,
    market: str,
    model_names: list[str] | None = None,
    echo: Optional[Callable[[Any], None]] = None,
    auto_yes: bool = False,
    echo_pretty: Optional[Callable[[Any], None]] = None,
) -> list[Model]:
    ...


def apply(
    path: Path,
    market: str,
    model_names: list[str] | str | None = None,
    echo: Optional[Callable[[Any], None]] = None,
    auto_yes: bool = False,
    echo_pretty: Optional[Callable[[Any], None]] = None,
) -> Model | list[Model]:
    echo = echo or print
    echo_pretty = echo_pretty or echo
    client = get_powerops_client()

    collection, config, models = _load_transform(
        market, path, client.cdf.config.project, echo, model_names or list(DEFAULT_MODELS)
    )
    collection.add(_create_bootstrap_finished_event(echo))

    _remove_non_existing_relationship_time_series_targets(client.cdf, models, collection, echo)

    summaries = {}
    for model in models:
        summaries.update(model.summary())

    echo("Models About to be uploaded")
    echo_pretty(summaries)
    ans = "y" if auto_yes else input("Continue? (y/n)")
    if ans.lower() == "y":
        # Step 3 - write bootstrap resources from diffs to CDF
        collection.write_to_cdf(
            client,
            config.settings.data_set_external_id,
            config.settings.overwrite_data,
        )
        echo("Resync written to CDF")
    else:
        echo("Aborting")

    return models[0] if len(models) == 1 else models


def _load_transform(
    market: str, path: Path, cdf_project: str, echo: Callable[[str], None], model_names: str | list[str] | None
) -> tuple[ResourceCollection, ReSyncConfig, list[Model]]:
    if isinstance(model_names, str):
        model_names = [model_names]
    elif model_names is None:
        model_names = list(DEFAULT_MODELS)
    elif isinstance(model_names, list) and model_names and isinstance(model_names[0], str):
        model_names = model_names
    else:
        raise ValueError(f"Invalid model_names type: {type(model_names)}")

    echo(f"Loading and transforming {', '.join(model_names)}")
    # Step 1 - configure and validate config
    config = ReSyncConfig.from_yamls(path, cdf_project)
    configure_debug_logging(config.settings.debug_level)
    # Step 2 - transform from config to CDF resources and preview diffs
    echo(
        f"Running resync for data set {config.settings.data_set_external_id} "
        f"in CDF project {config.settings.cdf_project}"
    )
    if invalid := set(model_names) - AVAILABLE_MODELS:
        raise ValueError(f"Invalid model names: {invalid}. Available models: {AVAILABLE_MODELS}")

    bootstrap_resources, models = transform(config, market, {MODEL_BY_NAME[model_name] for model_name in model_names})
    return bootstrap_resources, config, models


def _create_bootstrap_finished_event(echo: Callable[[str], None]) -> Event:
    """Creating a POWEROPS_BOOTSTRAP_FINISHED Event in CDF to signal that bootstrap scripts have been ran"""
    current_time = int(datetime.now(timezone.utc).timestamp() * 1000)  # in milliseconds
    event = Event(
        start_time=current_time,
        end_time=current_time,
        external_id=f"POWEROPS_BOOTSTRAP_FINISHED_{str(uuid4())}",
        type="POWEROPS_BOOTSTRAP_FINISHED",
        source="PowerOps bootstrap",
        description="Manual run of bootstrap scripts finished",
    )
    echo(f"Created status event '{event.external_id}'")

    return event


def _remove_non_existing_relationship_time_series_targets(
    client: CogniteClient, models: list[Model], collection: ResourceCollection, echo: Callable[[str], None]
) -> None:
    """Validates that all relationships in the collection have targets that exist in CDF"""
    to_delete = set()
    for model in models:
        if not isinstance(model, AssetModel):
            continue
        time_series = model.time_series()
        # retrieve_multiple fails if no time series are provided
        if not time_series:
            continue

        existing_time_series = client.time_series.retrieve_multiple(
            external_ids=list({t.external_id for t in time_series if t.external_id}), ignore_unknown_ids=True
        )
        existing_timeseries_ids = {ts.external_id: ts for ts in existing_time_series}
        missing_timeseries = {t.external_id for t in time_series if t.external_id not in existing_timeseries_ids}

        relationships = model.relationships()
        to_delete = {
            r.external_id
            for r in relationships
            if r.target_type and r.target_type.casefold() == "timeseries" and r.target_external_id in missing_timeseries
        }
        if to_delete:
            echo(
                f"WARNING: There are {len(to_delete)} relationships in {model.model_name} that have targets "
                "that do not exist in CDF. These relationships will not be created."
            )

        for external_id in to_delete:
            if external_id:
                collection.relationships.pop(external_id, None)


if __name__ == "__main__":
    demo_data = Path(__file__).parent.parent.parent.parent / "tests" / "test_unit" / "test_bootstrap" / "data" / "demo"

    apply(demo_data, "DayAhead", echo=print)



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from cognite.powerops.clients.data_classes._core import DomainModel, DomainModelApply, TypeList

if TYPE_CHECKING:
    from cognite.powerops.clients.data_classes._command_configs import CommandConfigApply
    from cognite.powerops.clients.data_classes._scenario_mappings import ScenarioMappingApply
    from cognite.powerops.clients.data_classes._scenario_templates import ScenarioTemplateApply

__all__ = ["Scenario", "ScenarioApply", "ScenarioList"]


class Scenario(DomainModel):
    space: ClassVar[str] = "power-ops"
    commands: Optional[str] = None
    mapping: Optional[str] = None
    name: Optional[str] = None
    template: Optional[str] = None


class ScenarioApply(DomainModelApply):
    space: ClassVar[str] = "power-ops"
    commands: Optional[Union["CommandConfigApply", str]] = Field(None, repr=False)
    mapping: Optional[Union["ScenarioMappingApply", str]] = Field(None, repr=False)
    name: Optional[str] = None
    template: Optional[Union["ScenarioTemplateApply", str]] = Field(None, repr=False)

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        source = dm.NodeOrEdgeData(
            source=dm.ContainerId("power-ops", "Scenario"),
            properties={
                "commands": {
                    "space": "power-ops",
                    "externalId": self.commands if isinstance(self.commands, str) else self.commands.external_id,
                },
                "mapping": {
                    "space": "power-ops",
                    "externalId": self.mapping if isinstance(self.mapping, str) else self.mapping.external_id,
                },
                "name": self.name,
                "template": {
                    "space": "power-ops",
                    "externalId": self.template if isinstance(self.template, str) else self.template.external_id,
                },
            },
        )
        sources.append(source)

        this_node = dm.NodeApply(
            space=self.space,
            external_id=self.external_id,
            existing_version=self.existing_version,
            sources=sources,
        )
        nodes = [this_node]
        edges = []
        cache.add(self.external_id)

        if isinstance(self.commands, DomainModelApply):
            instances = self.commands._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        if isinstance(self.mapping, DomainModelApply):
            instances = self.mapping._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        if isinstance(self.template, DomainModelApply):
            instances = self.template._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))


class ScenarioList(TypeList[Scenario]):
    _NODE = Scenario



import abc
import inspect
import json
import re
from hashlib import md5
from typing import Any, Type, TypeVar

from cognite.powerops.clients.data_classes._core import DomainModelApply


def special_case_handle_gate_number(name: str) -> None:
    """Must handle any gate numbers above 1 if found in other sources than YAML"""
    # TODO: extend to handle special case if needed
    if re.search(pattern=r"L[2-9]", string=name):
        print_warning(f"Potential gate {name} not in YAML!")


def print_warning(s: str) -> None:
    """Adds some nice colors to the printed text :)"""
    print(f"\033[91m[WARNING] {s}\033[0m")


def make_ext_id(arg: Any, class_: Type[DomainModelApply]) -> str:
    hash_value = md5()
    if isinstance(arg, (str, int, float, bool)):
        hash_value.update(str(arg).encode())
    elif isinstance(arg, (list, dict, tuple)):
        hash_value.update(json.dumps(arg).encode())
    return f"{class_.__name__.removesuffix('Apply')}__{hash_value.hexdigest()}"


T_Type = TypeVar("T_Type", bound=type)


def all_subclasses(base: T_Type) -> list[T_Type]:
    """Returns a list (without duplicates) of all subclasses of a given class, sorted on import-path-name.
    Ignores classes not part of the main library, e.g. subclasses part of tests.
    """
    return sorted(
        filter(
            lambda sub: str(sub).startswith("<class 'cognite.powerops"),
            set(base.__subclasses__()).union(s for c in base.__subclasses__() for s in all_subclasses(c)),
        ),
        key=str,
    )


def all_concrete_subclasses(base: T_Type) -> list[T_Type]:
    return [
        sub
        for sub in all_subclasses(base)
        if all(base is not abc.ABC for base in sub.__bases__) and not inspect.isabstract(sub)
    ]



from __future__ import annotations

from itertools import product
from typing import Type

from cognite.client import CogniteClient
from pydantic import Field

from cognite.powerops.clients.data_classes import (
    ScenarioTemplateApply,
    ValueTransformationApply,
    OutputContainerApply,
    ScenarioMappingApply,
)
import cognite.powerops.cogshop1.data_classes as cogshop_v1
from .base import DataModel, Model, T_Model
from .cdf_resources import CDFFile, CDFSequence
from cognite.powerops.clients.powerops_client import PowerOpsClient

ExternalID = str


class CogShopCore(Model):
    shop_files: list[CDFFile] = Field(default_factory=list)


class CogShopDataModel(CogShopCore, DataModel):
    scenario_templates: list[ScenarioTemplateApply] = Field(default_factory=list)
    base_mappings: dict[ExternalID, ScenarioMappingApply] = Field(default_factory=dict)
    output_definitions: dict[ExternalID, OutputContainerApply] = Field(default_factory=dict)
    value_transformations: dict[ExternalID, ValueTransformationApply] = Field(default_factory=dict)

    @classmethod
    def from_cdf(
        cls: Type[T_Model], client: PowerOpsClient, fetch_metadata: bool = True, fetch_content: bool = False
    ) -> T_Model:
        ...


class CogShop1Asset(CogShopCore, DataModel, protected_namespaces=()):
    model_templates: dict[ExternalID, cogshop_v1.ModelTemplateApply] = Field(default_factory=dict)
    base_mappings: list[CDFSequence] = Field(default_factory=list)
    output_definitions: list[CDFSequence] = Field(default_factory=list)

    @classmethod
    def from_cdf(
        cls, client: PowerOpsClient, data_set_id: int, fetch_metadata: bool = True, fetch_content: bool = False
    ) -> "CogShop1Asset":
        cog_shop = client.cog_shop1
        templates = cog_shop.model_templates.list(limit=-1)
        base_mapping_ids = list({mapping for t in templates for mapping in t.base_mappings})
        base_mappings = cog_shop.mappings.retrieve(base_mapping_ids)
        transformations_ids = list({t for m in base_mappings for t in m.transformations})
        transformations = cog_shop.transformations.retrieve(transformations_ids)
        file_ids = list({t.model for t in templates})
        files = cog_shop.file_refs.retrieve(file_ids)
        transformation_by_id = {t.external_id: t for t in transformations}
        mappings_by_id = {}
        readme_fields = {"created_time", "deleted_time", "last_updated_time", "version"}
        for mapping in base_mappings:
            data = mapping.model_dump(exclude=readme_fields)
            data["transformations"] = [
                transformation_by_id[t].model_dump(exclude=readme_fields) for t in data["transformations"]
            ]
            apply = cogshop_v1.MappingApply(**data)
            mappings_by_id[apply.external_id] = apply

        file_by_id = {f.external_id: f for f in files}
        model_templates = {}
        for template in templates:
            data = template.model_dump(exclude=readme_fields - {"version"})
            data["version"] = str(data["version"])
            if data["model"] in file_by_id:
                data["model"] = file_by_id[data["model"]].model_dump(exclude=readme_fields)
            data["base_mappings"] = [mappings_by_id[m] for m in data["base_mappings"]]
            apply = cogshop_v1.ModelTemplateApply(**data)
            model_templates[apply.external_id] = apply

        # There files and sequences are not linked to the model templates
        # (this should be done in the next version of Cogshop).
        watercourse_names = list({t.watercourse for t in model_templates.values()})
        sequence_ids = [
            f"SHOP_{name}{suffix}"
            for name, suffix in product(watercourse_names, ["_base_mapping", "_output_definition"])
        ]
        cdf_client: CogniteClient = client.cdf
        sequences = cdf_client.sequences.retrieve_multiple(external_ids=sequence_ids)

        base_mappings = [CDFSequence(sequence=s) for s in sequences if s.external_id.endswith("_base_mapping")]
        output_definitions = [
            CDFSequence(sequence=s) for s in sequences if s.external_id.endswith("_output_definition")
        ]

        files = cdf_client.files.list(
            limit=-1,
            source="PowerOps bootstrap",
            mime_type="text/plain",
        )
        shop_files = [CDFFile(meta=f) for f in files]

        return cls(
            model_templates=model_templates,
            base_mappings=base_mappings,
            output_definitions=output_definitions,
            shop_files=shop_files,
        )



from __future__ import annotations

from typing import Sequence, overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client._constants import INSTANCES_LIST_LIMIT_DEFAULT

from cognite.powerops.clients._api._core import TypeAPI
from cognite.powerops.clients.data_classes import NordPoolMarket, NordPoolMarketApply, NordPoolMarketList


class NordPoolMarketsAPI(TypeAPI[NordPoolMarket, NordPoolMarketApply, NordPoolMarketList]):
    def __init__(self, client: CogniteClient):
        super().__init__(
            client=client,
            sources=dm.ViewId("power-ops", "NordPoolMarket", "88c86032b9ac9c"),
            class_type=NordPoolMarket,
            class_apply_type=NordPoolMarketApply,
            class_list=NordPoolMarketList,
        )

    def apply(self, nord_pool_market: NordPoolMarketApply, replace: bool = False) -> dm.InstancesApplyResult:
        instances = nord_pool_market.to_instances_apply()
        return self._client.data_modeling.instances.apply(nodes=instances.nodes, edges=instances.edges, replace=replace)

    def delete(self, external_id: str | Sequence[str]) -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(NordPoolMarketApply.space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(NordPoolMarketApply.space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> NordPoolMarket:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> NordPoolMarketList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> NordPoolMarket | NordPoolMarketList:
        if isinstance(external_id, str):
            return self._retrieve((self.sources.space, external_id))
        else:
            return self._retrieve([(self.sources.space, ext_id) for ext_id in external_id])

    def list(self, limit: int = INSTANCES_LIST_LIMIT_DEFAULT) -> NordPoolMarketList:
        return self._list(limit=limit)



from __future__ import annotations

from collections import defaultdict
from typing import Dict, List, Sequence, Tuple, overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client._constants import INSTANCES_LIST_LIMIT_DEFAULT

from cognite.powerops.cogshop1._api._core import TypeAPI
from cognite.powerops.cogshop1.data_classes import ModelTemplate, ModelTemplateApply, ModelTemplateList


class ModelTemplateBaseMappingsAPI:
    def __init__(self, client: CogniteClient):
        self._client = client

    def retrieve(self, external_id: str | Sequence[str]) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "cogShop", "externalId": "ModelTemplate.baseMappings"},
        )
        if isinstance(external_id, str):
            is_model_template = f.Equals(
                ["edge", "startNode"],
                {"space": "cogShop", "externalId": external_id},
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_model_template)
            )

        else:
            is_model_templates = f.In(
                ["edge", "startNode"],
                [{"space": "cogShop", "externalId": ext_id} for ext_id in external_id],
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_model_templates)
            )

    def list(self, limit=INSTANCES_LIST_LIMIT_DEFAULT) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "cogShop", "externalId": "ModelTemplate.baseMappings"},
        )
        return self._client.data_modeling.instances.list("edge", limit=limit, filter=is_edge_type)


class ModelTemplatesAPI(TypeAPI[ModelTemplate, ModelTemplateApply, ModelTemplateList]):
    def __init__(self, client: CogniteClient):
        super().__init__(
            client=client,
            sources=dm.ViewId("cogShop", "ModelTemplate", "d5ec6bd344c88d"),
            class_type=ModelTemplate,
            class_apply_type=ModelTemplateApply,
            class_list=ModelTemplateList,
        )
        self.base_mappings = ModelTemplateBaseMappingsAPI(client)

    def apply(self, model_template: ModelTemplateApply, replace: bool = False) -> dm.InstancesApplyResult:
        instances = model_template.to_instances_apply()
        return self._client.data_modeling.instances.apply(nodes=instances.nodes, edges=instances.edges, replace=replace)

    def delete(self, external_id: str | Sequence[str]) -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(ModelTemplateApply.space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(ModelTemplateApply.space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> ModelTemplate:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> ModelTemplateList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> ModelTemplate | ModelTemplateList:
        if isinstance(external_id, str):
            model_template = self._retrieve((self.sources.space, external_id))

            base_mapping_edges = self.base_mappings.retrieve(external_id)
            model_template.base_mappings = [edge.end_node.external_id for edge in base_mapping_edges]

            return model_template
        else:
            model_templates = self._retrieve([(self.sources.space, ext_id) for ext_id in external_id])

            base_mapping_edges = self.base_mappings.retrieve(external_id)
            self._set_base_mappings(model_templates, base_mapping_edges)

            return model_templates

    def list(self, limit: int = INSTANCES_LIST_LIMIT_DEFAULT) -> ModelTemplateList:
        model_templates = self._list(limit=limit)

        base_mapping_edges = self.base_mappings.list(limit=-1)
        self._set_base_mappings(model_templates, base_mapping_edges)

        return model_templates

    @staticmethod
    def _set_base_mappings(model_templates: Sequence[ModelTemplate], base_mapping_edges: Sequence[dm.Edge]):
        edges_by_start_node: Dict[Tuple, List] = defaultdict(list)
        for edge in base_mapping_edges:
            edges_by_start_node[edge.start_node.as_tuple()].append(edge)

        for model_template in model_templates:
            node_id = model_template.id_tuple()
            if node_id in edges_by_start_node:
                model_template.base_mappings = [edge.end_node.external_id for edge in edges_by_start_node[node_id]]



from .market import MarketModel
from .production import ProductionModel
from .production_dm import ProductionModelDM
from .cogshop import CogShopDataModel, CogShop1Asset
from .market_dm import BenchmarkMarketDataModel, DayAheadMarketDataModel, RKOMMarketDataModel

__all__ = [
    "ProductionModel",
    "ProductionModelDM",
    "MarketModel",
    "CogShopDataModel",
    "CogShop1Asset",
    "BenchmarkMarketDataModel",
    "DayAheadMarketDataModel",
    "RKOMMarketDataModel",
]



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from cognite.powerops.clients.data_classes._core import DomainModel, DomainModelApply, TypeList

if TYPE_CHECKING:
    from cognite.powerops.clients.data_classes._date_transformations import DateTransformationApply
    from cognite.powerops.clients.data_classes._reserve_scenarios import ReserveScenarioApply
    from cognite.powerops.clients.data_classes._rkom_markets import RKOMMarketApply
    from cognite.powerops.clients.data_classes._scenario_mappings import ScenarioMappingApply

__all__ = ["RKOMBid", "RKOMBidApply", "RKOMBidList"]


class RKOMBid(DomainModel):
    space: ClassVar[str] = "power-ops"
    date: list[str] = []
    market: Optional[str] = None
    method: Optional[str] = None
    minimum_price: Optional[float] = Field(None, alias="minimumPrice")
    name: Optional[str] = None
    price_premium: Optional[float] = Field(None, alias="pricePremium")
    price_scenarios: list[str] = Field([], alias="priceScenarios")
    reserve_scenarios: list[str] = Field([], alias="reserveScenarios")
    watercourse: Optional[str] = None


class RKOMBidApply(DomainModelApply):
    space: ClassVar[str] = "power-ops"
    date: list[Union["DateTransformationApply", str]] = Field(default_factory=list, repr=False)
    market: Optional[Union["RKOMMarketApply", str]] = Field(None, repr=False)
    method: Optional[str] = None
    minimum_price: Optional[float] = None
    name: Optional[str] = None
    price_premium: Optional[float] = None
    price_scenarios: list[Union["ScenarioMappingApply", str]] = Field(default_factory=list, repr=False)
    reserve_scenarios: list[Union["ReserveScenarioApply", str]] = Field(default_factory=list, repr=False)
    watercourse: Optional[str] = None

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        source = dm.NodeOrEdgeData(
            source=dm.ContainerId("power-ops", "RKOMBid"),
            properties={
                "market": {
                    "space": "power-ops",
                    "externalId": self.market if isinstance(self.market, str) else self.market.external_id,
                },
                "method": self.method,
                "minimumPrice": self.minimum_price,
                "name": self.name,
                "pricePremium": self.price_premium,
                "watercourse": self.watercourse,
            },
        )
        sources.append(source)

        this_node = dm.NodeApply(
            space=self.space,
            external_id=self.external_id,
            existing_version=self.existing_version,
            sources=sources,
        )
        nodes = [this_node]
        edges = []
        cache.add(self.external_id)

        for date in self.date:
            edge = self._create_date_edge(date)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(date, DomainModelApply):
                instances = date._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        for price_scenario in self.price_scenarios:
            edge = self._create_price_scenario_edge(price_scenario)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(price_scenario, DomainModelApply):
                instances = price_scenario._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        for reserve_scenario in self.reserve_scenarios:
            edge = self._create_reserve_scenario_edge(reserve_scenario)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(reserve_scenario, DomainModelApply):
                instances = reserve_scenario._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        if isinstance(self.market, DomainModelApply):
            instances = self.market._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))

    def _create_date_edge(self, date: Union[str, "DateTransformationApply"]) -> dm.EdgeApply:
        if isinstance(date, str):
            end_node_ext_id = date
        elif isinstance(date, DomainModelApply):
            end_node_ext_id = date.external_id
        else:
            raise TypeError(f"Expected str or DateTransformationApply, got {type(date)}")

        return dm.EdgeApply(
            space="power-ops",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("power-ops", "RKOMBid.date"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("power-ops", end_node_ext_id),
        )

    def _create_price_scenario_edge(self, price_scenario: Union[str, "ScenarioMappingApply"]) -> dm.EdgeApply:
        if isinstance(price_scenario, str):
            end_node_ext_id = price_scenario
        elif isinstance(price_scenario, DomainModelApply):
            end_node_ext_id = price_scenario.external_id
        else:
            raise TypeError(f"Expected str or ScenarioMappingApply, got {type(price_scenario)}")

        return dm.EdgeApply(
            space="power-ops",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("power-ops", "RKOMBid.priceScenarios"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("power-ops", end_node_ext_id),
        )

    def _create_reserve_scenario_edge(self, reserve_scenario: Union[str, "ReserveScenarioApply"]) -> dm.EdgeApply:
        if isinstance(reserve_scenario, str):
            end_node_ext_id = reserve_scenario
        elif isinstance(reserve_scenario, DomainModelApply):
            end_node_ext_id = reserve_scenario.external_id
        else:
            raise TypeError(f"Expected str or ReserveScenarioApply, got {type(reserve_scenario)}")

        return dm.EdgeApply(
            space="power-ops",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("power-ops", "RKOMBid.reserveScenarios"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("power-ops", end_node_ext_id),
        )


class RKOMBidList(TypeList[RKOMBid]):
    _NODE = RKOMBid



from __future__ import annotations

from pathlib import Path
from typing import ClassVar, Optional, Union

from cognite.client.data_classes import Asset, TimeSeries
from pydantic import ConfigDict, Field, field_validator

from cognite.powerops.cdf_labels import AssetLabel
from cognite.powerops.resync.models.base import AssetModel, AssetType, NonAssetType
from cognite.powerops.resync.models.cdf_resources import CDFSequence
from cognite.powerops.resync.models.helpers import isinstance_list
from cognite.powerops.resync.utils.serializer import try_load_dict


class Generator(AssetType):
    parent_external_id: ClassVar[str] = "generators"
    label: ClassVar[Union[AssetLabel, str]] = AssetLabel.GENERATOR
    p_min: float = 0.0
    penstock: str = ""
    startcost: float = 0.0
    start_stop_cost_time_series: Optional[TimeSeries] = None
    is_available_time_series: Optional[TimeSeries] = None
    generator_efficiency_curve: Optional[CDFSequence] = None
    turbine_efficiency_curve: Optional[CDFSequence] = None


class Reservoir(AssetType):
    parent_external_id: ClassVar[str] = "reservoirs"
    label: ClassVar[Union[AssetLabel, str]] = AssetLabel.RESERVOIR
    display_name: str = ""
    ordering: str = ""


class Plant(AssetType):
    parent_external_id: ClassVar[str] = "plants"
    label: ClassVar[Union[AssetLabel, str]] = AssetLabel.PLANT
    display_name: str = ""
    ordering: str = ""
    head_loss_factor: float = 0.0
    outlet_level: float = 0.0
    p_min: float = 0.0
    p_max: float = 0.0
    penstock_head_loss_factors: dict = Field(default_factory=dict)
    generators: list[Generator] = Field(default_factory=list)
    inlet_reservoir: Optional[Reservoir] = None
    p_min_time_series: Optional[TimeSeries] = None
    p_max_time_series: Optional[TimeSeries] = None
    water_value_time_series: Optional[TimeSeries] = None
    feeding_fee_time_series: Optional[TimeSeries] = None
    outlet_level_time_series: Optional[TimeSeries] = None
    inlet_level_time_series: Optional[TimeSeries] = None
    head_direct_time_series: Optional[TimeSeries] = None

    @field_validator("penstock_head_loss_factors", mode="before")
    def parse_str(cls, value) -> dict:
        return try_load_dict(value)


class WaterCourseShop(NonAssetType):
    penalty_limit: str


class Watercourse(AssetType):
    model_config: ClassVar[ConfigDict] = ConfigDict(protected_namespaces=tuple())
    parent_external_id: ClassVar[str] = "watercourses"
    label: ClassVar[Union[AssetLabel, str]] = AssetLabel.WATERCOURSE
    shop: WaterCourseShop
    config_version: Optional[str] = Field("", exclude=True)
    model_file: Optional[Path] = Field(None, exclude=True)
    processed_model_file: Optional[Path] = Field(None, exclude=True)
    plants: list[Plant]
    production_obligation_time_series: list[TimeSeries] = Field(default_factory=list)

    @field_validator("production_obligation_time_series", mode="before")
    def none_to_empty_list(cls, value) -> list[TimeSeries]:
        if value is None or (isinstance(value, list) and value and value[0] is None):
            return []
        return value


class PriceArea(AssetType):
    parent_external_id: ClassVar[str] = "price_areas"
    label: ClassVar[Union[AssetLabel, str]] = AssetLabel.PRICE_AREA
    dayahead_price_time_series: Optional[TimeSeries] = None
    plants: list[Plant] = Field(default_factory=list)
    watercourses: list[Watercourse] = Field(default_factory=list)


class ProductionModel(AssetModel):
    root_asset: ClassVar[Asset] = Asset(external_id="power_ops", name="PowerOps")
    plants: list[Plant] = Field(default_factory=list)
    generators: list[Generator] = Field(default_factory=list)
    reservoirs: list[Reservoir] = Field(default_factory=list)
    watercourses: list[Watercourse] = Field(default_factory=list)
    price_areas: list[PriceArea] = Field(default_factory=list)

    def _prepare_for_diff(self: ProductionModel) -> dict:
        clone = self.model_copy(deep=True)

        for model_field in clone.model_fields:
            field_value = getattr(clone, model_field)
            if isinstance_list(field_value, AssetType):
                # Sort the asset types to have comparable order for diff
                _sorted = sorted(field_value, key=lambda x: x.external_id)
                # Prepare each asset type for diff
                _prepared = map(lambda x: x._asset_type_prepare_for_diff(), _sorted)
                setattr(clone, model_field, list(_prepared))
            elif isinstance(field_value, AssetType):
                field_value._asset_type_prepare_for_diff()
        # Some fields are have been set to their external_id which gives a warning we can ignore
        return clone.model_dump(warnings=False)



from dataclasses import dataclass
from pathlib import Path

from cognite.client.data_classes.data_modeling import DataModelId

_POWEROPS_ROOT = Path(__file__).parent
GRAPHQL_SCHEMAS = _POWEROPS_ROOT / "data_models"
_SHARED_TYPES = GRAPHQL_SCHEMAS / "shared"


@dataclass
class PowerOpsModel:
    name: str
    description: str
    graphql_file: Path
    id_: DataModelId
    extra_types: list[Path] = None

    @property
    def graphql(self) -> str:
        graphql = self.graphql_file.read_text()
        if self.extra_types:
            graphql = "\n\n".join([extra.read_text() for extra in self.extra_types] + [graphql])

        return graphql


_SPACE = "power-ops"


class _ExtraTypes:
    input_timeseries_mapping = _SHARED_TYPES / "input_timeseries_mapping.graphql"
    value_transformation = _SHARED_TYPES / "value_transformation.graphql"
    date_transformation = _SHARED_TYPES / "date_transformation.graphql"
    scenario_mapping = _SHARED_TYPES / "scenario_mapping.graphql"
    shop_transformation = _SHARED_TYPES / "shop_transformation.graphql"
    nordpooL_market = _SHARED_TYPES / "nordpool_market.graphql"


MODEL_BY_NAME: dict[str, PowerOpsModel] = {
    "production": PowerOpsModel(
        name="Production",
        description="The production model describes the physical assets such as watercourses, "
        "plants, and generators located in a price area.",
        graphql_file=GRAPHQL_SCHEMAS / "production.graphql",
        id_=DataModelId(_SPACE, "production", "1"),
    ),
    "cogshop": PowerOpsModel(
        name="CogShop",
        description="The CogShop model describes the interaction between Cognite Data Fusion (CDF) and "
        "Sintef's SHOP algorithm. The scenario is used to determine which SHOP runs are"
        "executed daily and configuration of those SHOP runs.",
        graphql_file=GRAPHQL_SCHEMAS / "cogshop.graphql",
        id_=DataModelId(_SPACE, "cogshop", "1"),
        extra_types=[
            _ExtraTypes.value_transformation,
            _ExtraTypes.input_timeseries_mapping,
            _ExtraTypes.scenario_mapping,
        ],
    ),
    "dayahead": PowerOpsModel(
        name="DayAhead",
        description="The DayAhead model describes the day-ahead market.",
        graphql_file=GRAPHQL_SCHEMAS / "dayahead-market.graphql",
        id_=DataModelId(_SPACE, "dayaheadMarket", "1"),
        extra_types=[
            _ExtraTypes.value_transformation,
            _ExtraTypes.date_transformation,
            _ExtraTypes.input_timeseries_mapping,
            _ExtraTypes.scenario_mapping,
            _ExtraTypes.shop_transformation,
            _ExtraTypes.nordpooL_market,
        ],
    ),
    "rkom": PowerOpsModel(
        name="RKOMMarket",
        description="The RKOM market is a balancing market",
        graphql_file=GRAPHQL_SCHEMAS / "rkom-market.graphql",
        id_=DataModelId(_SPACE, "rkomMarket", "1"),
        extra_types=[
            _ExtraTypes.value_transformation,
            _ExtraTypes.date_transformation,
            _ExtraTypes.input_timeseries_mapping,
            _ExtraTypes.scenario_mapping,
            _ExtraTypes.shop_transformation,
        ],
    ),
    "benchmark": PowerOpsModel(
        name="Benchmark",
        description="The Benchmark model is used for benchmarking different bid processes.",
        graphql_file=GRAPHQL_SCHEMAS / "benchmark-market.graphql",
        id_=DataModelId(_SPACE, "benchmarkMarket", "1"),
        extra_types=[
            _ExtraTypes.value_transformation,
            _ExtraTypes.date_transformation,
            _ExtraTypes.input_timeseries_mapping,
            _ExtraTypes.scenario_mapping,
            _ExtraTypes.shop_transformation,
            _ExtraTypes.nordpooL_market,
        ],
    ),
}



from __future__ import annotations

import json
import logging
from datetime import datetime
from hashlib import md5
from pathlib import Path

import pandas as pd
import yaml
from cognite.client.data_classes import FileMetadata, Sequence

from cognite.powerops.clients.data_classes import (
    OutputMappingApply,
    ScenarioTemplateApply,
    ValueTransformationApply,
    OutputContainerApply,
    ScenarioMappingApply,
)
from cognite.powerops.resync.config.cogshop.shop_file_config import ShopFileConfig
from cognite.powerops.resync.config.production.watercourse import WatercourseConfig
from cognite.powerops.resync.config.resync_config import CogShopConfig
from cognite.powerops.resync.models import cogshop
from cognite.powerops.resync.models.cdf_resources import CDFSequence
from cognite.powerops.resync.models.cogshop import CogShopDataModel
from cognite.powerops.resync.models.production import Watercourse
from cognite.powerops.resync.utils.common import make_ext_id
from cognite.powerops.resync.utils.serializer import load_yaml
import cognite.powerops.cogshop1.data_classes as cogshop_v1

from ._to_instances import _to_input_timeseries_mapping

logger = logging.getLogger(__name__)


def to_cogshop_data_model(
    config: CogShopConfig, watercourse_configs: list[WatercourseConfig], shop_version: str
) -> CogShopDataModel:
    model = CogShopDataModel()

    model.shop_files.extend(_to_shop_files(config.watercourses_shop))

    # TODO Fix the assumption that timeseries mappings and watercourses are in the same order
    for watercourse, mapping in zip(watercourse_configs, config.time_series_mappings):
        model_file = _to_shop_model_file(watercourse.name, watercourse.yaml_raw_path, watercourse.yaml_processed_path)
        model.shop_files.append(model_file)

        ##### Output Definitions #####
        # Only default mapping is used
        output_definitions = [
            OutputMappingApply(
                external_id=make_ext_id(values, class_=OutputMappingApply),
                shop_object_type=values[0],
                shop_attribute_name=values[1],
                cdf_attribute_name=values[2],
                unit=values[3],
                is_step=values[4],
            )
            for values in [
                ("market", "sale_price", "price", "EUR/MWh", True),
                ("market", "sale", "sales", "MWh", True),
                ("plant", "production", "production", "MW", True),
                ("plant", "consumption", "consumption", "MW", True),
                ("reservoir", "water_value_global_result", "water_value", "EUR/Mm3", True),
                ("reservoir", "energy_conversion_factor", "energy_conversion_factor", "MWh/Mm3", True),
            ]
        ]
        external_id = f"SHOP_{watercourse.name.replace(' ', '_')}_output_definition"
        output_container = OutputContainerApply(
            external_id=external_id,
            name=external_id.replace("_", " "),
            watercourse=watercourse.name,
            shop_type="output_definition",
            mappings=output_definitions,
        )

        model.output_definitions[external_id] = output_container

        ##### Base Mapping #####
        base_mappings = []
        transformations = {}
        for entry in mapping:
            base_mapping = _to_input_timeseries_mapping(entry)
            base_mappings.append(base_mapping)

            transformations.update(
                {t.external_id: t for t in base_mapping.transformations if isinstance(t, ValueTransformationApply)}
            )
        scenario_mapping = ScenarioMappingApply(
            external_id=f"SHOP_{watercourse.name}_base_mapping",
            watercourse=watercourse.name,
            shop_type="base_mapping",
            mapping_override=base_mappings,
        )

        model.value_transformations.update(transformations)

        model.scenario_templates.append(
            ScenarioTemplateApply(
                external_id=f"{ScenarioTemplateApply.__name__.removesuffix('Apply')}_{watercourse.name}",
                model=model_file.external_id,
                shop_version=shop_version,
                template_version="1",
                output_definitions=output_container,
                shop_files=[f.external_id for f in model.shop_files],
                watercourse=watercourse.name,
                base_mapping=scenario_mapping,
            )
        )

    return model


def to_cogshop_asset_model(
    config: CogShopConfig, watercourses: list[Watercourse], shop_version: str
) -> cogshop.CogShop1Asset:
    model = cogshop.CogShop1Asset()

    model.shop_files.extend(_to_shop_files(config.watercourses_shop))

    # TODO Fix the assumption that timeseries mappings and watercourses are in the same order
    for watercourse, mapping in zip(watercourses, config.time_series_mappings):
        model_file = _to_shop_model_file(watercourse.name, watercourse.model_file, watercourse.processed_model_file)
        model.shop_files.append(model_file)

        ##### Output definition #####
        external_id = f"SHOP_{watercourse.name.replace(' ', '_')}_output_definition"

        sequence = Sequence(
            name=external_id.replace("_", " "),
            description="Defining which SHOP results to output to CDF (as time series)",
            external_id=external_id,
            columns=[
                {"valueType": "STRING", "externalId": "shop_object_type"},
                {"valueType": "STRING", "externalId": "shop_attribute_name"},
                {"valueType": "STRING", "externalId": "cdf_attribute_name"},
                {"valueType": "STRING", "externalId": "unit"},
                {"valueType": "STRING", "externalId": "is_step"},
            ],
            metadata={
                "shop:watercourse": watercourse.name,
                "shop:type": "output_definition",
            },
        )
        # Only default mapping is used
        df = pd.DataFrame(
            [
                ("market", "sale_price", "price", "EUR/MWh", "True"),
                ("market", "sale", "sales", "MWh", "True"),
                ("plant", "production", "production", "MW", "True"),
                ("plant", "consumption", "consumption", "MW", "True"),
                ("reservoir", "water_value_global_result", "water_value", "EUR/Mm3", "True"),
                ("reservoir", "energy_conversion_factor", "energy_conversion_factor", "MWh/Mm3", "True"),
            ],
            columns=[c["externalId"] for c in sequence.columns],
        )

        output_definition = CDFSequence(
            sequence=sequence,
            content=df,
        )

        model.output_definitions.append(output_definition)

        ##### Base Mapping #####
        external_id = f"SHOP_{watercourse.name}_base_mapping"
        sequence = Sequence(
            name=external_id.replace("_", " "),
            external_id=external_id,
            description="Mapping between SHOP paths and CDF TimeSeries",
            columns=mapping.column_definitions,
            metadata={
                "shop:watercourse": watercourse.name,
                "shop:type": "base_mapping",
            },
        )
        output_definition = CDFSequence(
            sequence=sequence,
            content=mapping.to_dataframe(),
        )
        model.base_mappings.append(output_definition)

        ### Model Template ###
        def make_ext_id(watercourse_name: str, *args: str) -> str:
            hash_value = md5(watercourse_name.encode())
            for arg in args:
                hash_value.update(arg.encode())
            return f"Tr__{hash_value.hexdigest()}"

        model_template = cogshop_v1.ModelTemplateApply(
            external_id=f"ModelTemplate_{watercourse.name}",
            version="1",
            shop_version=shop_version,
            watercourse=watercourse.name,
            model=cogshop_v1.FileRefApply(
                external_id=f"ModelTemplate_{watercourse.name}__FileRef_model",
                type="case",
                file_external_id=model_file.external_id,
            ),
            base_mappings=[
                cogshop_v1.MappingApply(
                    external_id=f"BM__{watercourse.name}__{row.shop_model_path}",
                    path=row.shop_model_path,
                    timeseries_external_id=row.time_series_external_id,
                    transformations=[
                        cogshop_v1.TransformationApply(
                            external_id=make_ext_id(
                                watercourse.name,
                                row.shop_model_path,
                                transformation.transformation.name,
                                json.dumps(transformation.kwargs or {}),
                            ),
                            method=transformation.transformation.name,
                            arguments=json.dumps(transformation.kwargs or {}),
                        )
                        for transformation in (row.transformations or [])
                    ],
                    retrieve=row.retrieve.name if row.retrieve else None,
                    aggregation=row.aggregation.name if row.aggregation else None,
                )
                for row in mapping
            ],
        )
        model.model_templates[model_template.external_id] = model_template

    return model


def _to_shop_model_file(watercourse_name, model_file: Path, processed_model_file: Path) -> cogshop.CDFFile:
    processed_model = _to_model_without_timeseries(
        yaml_raw_path=str(model_file),
        # Todo Move this hardcoded configuration to a config file
        non_time_series_attributes_to_remove={"reservoir.start_vol", "generator.initial_state"},
        encoding="utf-8",
    )
    yaml_content = yaml.safe_dump(processed_model, allow_unicode=True, sort_keys=False)
    file_content = yaml_content.encode("utf-8")
    external_id = f"SHOP_{watercourse_name}_{processed_model_file.stem}"
    return _create_shop_file(
        file_content,
        external_id,
        {
            "shop:type": "case",
            "shop:watercourse": watercourse_name,
            "shop:file_name": processed_model_file.stem,
        },
    )


def _to_shop_files(watercourses_shop: list[ShopFileConfig]) -> list[cogshop.CDFFile]:
    shop_files = []
    for shop_config in watercourses_shop:
        file_content = Path(shop_config.file_path).read_bytes()
        shop_file = _create_shop_file(
            file_content,
            shop_config.external_id,
            {
                "shop:type": shop_config.cogshop_file_type,
                "shop:watercourse": shop_config.watercourse_name,
                "shop:file_name": shop_config.file_path.stem,
            },
        )
        shop_files.append(shop_file)
    return shop_files


def _create_shop_file(file_content: bytes, external_id: str, metadata: dict[str, str]) -> cogshop.CDFFile:
    metadata["md5_hash"] = md5(file_content.replace(b"\r\n", b"\n")).hexdigest()
    return cogshop.CDFFile(
        meta=FileMetadata(
            external_id=external_id,
            name=external_id,
            metadata=metadata,
            source="PowerOps bootstrap",
            mime_type="text/plain",
        ),
        content=file_content,
    )


def _is_time_series(shop_attribute_value) -> bool:
    return isinstance(shop_attribute_value, dict) and isinstance(list(shop_attribute_value)[0], datetime)


def _to_model_without_timeseries(
    yaml_raw_path: str,
    non_time_series_attributes_to_remove: set[str] | None = None,
    encoding=None,
) -> dict:
    non_time_series_attributes_to_remove = non_time_series_attributes_to_remove or set()

    model_incl_time_and_connections = load_yaml(Path(yaml_raw_path), encoding=encoding, clean_data=True)

    model = model_incl_time_and_connections["model"]

    for object_type, objects in model.items():
        for object_attributes in objects.values():
            for attribute_name, attribute_data in list(
                object_attributes.items()
            ):  # Needs list() since we pop while iterating over the dict
                if f"{object_type}.{attribute_name}" in non_time_series_attributes_to_remove:
                    object_attributes.pop(attribute_name)

                elif _is_time_series(attribute_data):
                    object_attributes.pop(attribute_name)

    model_dict = {"model": model}

    # TODO: remove this when standardizing input SHOP file structure (model vs model + connections)
    if "connections" in model_incl_time_and_connections:
        model_dict["connections"] = model_incl_time_and_connections["connections"]

    return model_dict



from __future__ import annotations

from typing import Optional

from pydantic import BaseModel, field_validator

from cognite.powerops.resync.config.shared import ExternalId

GeneratorName = str


class Generator(BaseModel):
    name: GeneratorName
    penstock: str
    startcost: float
    p_min: float

    start_stop_cost_time_series: Optional[ExternalId] = None  # external ID of time series with values in m
    is_available_time_series: Optional[ExternalId] = None  # external ID of boolean time series

    @property
    def external_id(self) -> ExternalId:
        return f"generator_{self.name}"


class GeneratorTimeSeriesMapping(BaseModel):
    generator_name: str
    start_stop_cost: Optional[ExternalId] = None
    is_available: Optional[ExternalId] = None

    @field_validator("start_stop_cost", "is_available", mode="before")
    def parset_number_to_string(cls, value):
        return str(value) if isinstance(value, (int, float)) else value



from __future__ import annotations

import json
import re
import string
import warnings
from functools import lru_cache
from io import StringIO
from pathlib import Path
from typing import Any
from yaml import safe_dump, safe_load
from cognite.client.utils._text import to_camel_case

# � character is used to represent unrecognizable characters in utf-8.
UNRECOGNIZABLE_CHARACTER = "�"
VALID_CHARACTERS = set(
    string.ascii_lowercase
    + string.ascii_uppercase
    + string.digits
    + UNRECOGNIZABLE_CHARACTER
    + string.punctuation
    + string.whitespace
    + "æøåÆØÅ"
)

_READ_ONLY_FIELDS = ["created_time", "last_updated_time", "uploaded_time", "data_set_id", "id", "parent_id", "root_id"]


def try_load_list(value: str | Any) -> Any | list[Any]:
    if isinstance(value, str):
        try:
            return json.loads(value)
        except json.JSONDecodeError:
            return []
    return value


def try_load_dict(value: str | Any) -> Any | dict[str, Any]:
    if isinstance(value, str):
        try:
            return json.loads(value)
        except json.JSONDecodeError:
            return {}
    return value


def remove_read_only_fields(cdf_resource: dict[str, Any], remove_empty: bool = True) -> dict[str, Any]:
    for field in _READ_ONLY_FIELDS:
        cdf_resource.pop(field, None)
        cdf_resource.pop(to_camel_case(field), None)
    if remove_empty:
        for field in list(cdf_resource):
            if not cdf_resource[field]:
                cdf_resource.pop(field)
    return cdf_resource


def _validate(yaml_path: Path):
    if yaml_path.suffix not in {".yaml", ".yml"}:
        raise ValueError(f"File {yaml_path.name} not a valid yaml {yaml_path.suffix}")


@lru_cache
def load_yaml(yaml_path: Path, encoding="utf-8", clean_data: bool = False) -> dict:
    _validate(yaml_path)
    # The Windows Cpython implementation seems to guess if encoding is not explicitly set
    # This turns out to be a problem as it guesses wrong, which is not the case for Unix systems.
    data = Path(yaml_path).read_text(encoding=encoding)

    if clean_data and (invalid_characters := (set(data) - VALID_CHARACTERS)):
        data = re.sub(rf"[{'|'.join(invalid_characters)}]", UNRECOGNIZABLE_CHARACTER, data)
        warnings.warn(
            f"File {yaml_path.parent}/{yaml_path.name} contains invalid characters: {', '.join(invalid_characters)}"
        )
    return safe_load(StringIO(data))


# Having a single yaml dump function makes it easy to look up all places YaMLs are dumped.
# It also enables one place to set the default encoding.
def dump_yaml(data: dict, yaml_path: Path, encoding="utf-8"):
    _validate(yaml_path)
    with open(yaml_path, "w", encoding=encoding) as stream:
        safe_dump(data, stream)



from __future__ import annotations

from collections import defaultdict
from typing import Dict, List, Sequence, Tuple, overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client._constants import INSTANCES_LIST_LIMIT_DEFAULT

from cognite.powerops.cogshop1._api._core import TypeAPI
from cognite.powerops.cogshop1.data_classes import Scenario, ScenarioApply, ScenarioList


class ScenarioExtraFilesAPI:
    def __init__(self, client: CogniteClient):
        self._client = client

    def retrieve(self, external_id: str | Sequence[str]) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "cogShop", "externalId": "Scenario.extraFiles"},
        )
        if isinstance(external_id, str):
            is_scenario = f.Equals(
                ["edge", "startNode"],
                {"space": "cogShop", "externalId": external_id},
            )
            return self._client.data_modeling.instances.list("edge", limit=-1, filter=f.And(is_edge_type, is_scenario))

        else:
            is_scenarios = f.In(
                ["edge", "startNode"],
                [{"space": "cogShop", "externalId": ext_id} for ext_id in external_id],
            )
            return self._client.data_modeling.instances.list("edge", limit=-1, filter=f.And(is_edge_type, is_scenarios))

    def list(self, limit=INSTANCES_LIST_LIMIT_DEFAULT) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "cogShop", "externalId": "Scenario.extraFiles"},
        )
        return self._client.data_modeling.instances.list("edge", limit=limit, filter=is_edge_type)


class ScenarioMappingsOverridesAPI:
    def __init__(self, client: CogniteClient):
        self._client = client

    def retrieve(self, external_id: str | Sequence[str]) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "cogShop", "externalId": "Scenario.mappingsOverride"},
        )
        if isinstance(external_id, str):
            is_scenario = f.Equals(
                ["edge", "startNode"],
                {"space": "cogShop", "externalId": external_id},
            )
            return self._client.data_modeling.instances.list("edge", limit=-1, filter=f.And(is_edge_type, is_scenario))

        else:
            is_scenarios = f.In(
                ["edge", "startNode"],
                [{"space": "cogShop", "externalId": ext_id} for ext_id in external_id],
            )
            return self._client.data_modeling.instances.list("edge", limit=-1, filter=f.And(is_edge_type, is_scenarios))

    def list(self, limit=INSTANCES_LIST_LIMIT_DEFAULT) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "cogShop", "externalId": "Scenario.mappingsOverride"},
        )
        return self._client.data_modeling.instances.list("edge", limit=limit, filter=is_edge_type)


class ScenariosAPI(TypeAPI[Scenario, ScenarioApply, ScenarioList]):
    def __init__(self, client: CogniteClient):
        super().__init__(
            client=client,
            sources=dm.ViewId("cogShop", "Scenario", "3f195f52332116"),
            class_type=Scenario,
            class_apply_type=ScenarioApply,
            class_list=ScenarioList,
        )
        self.extra_files = ScenarioExtraFilesAPI(client)
        self.mappings_overrides = ScenarioMappingsOverridesAPI(client)

    def apply(self, scenario: ScenarioApply, replace: bool = False) -> dm.InstancesApplyResult:
        instances = scenario.to_instances_apply()
        return self._client.data_modeling.instances.apply(nodes=instances.nodes, edges=instances.edges, replace=replace)

    def delete(self, external_id: str | Sequence[str]) -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(ScenarioApply.space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(ScenarioApply.space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> Scenario:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> ScenarioList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> Scenario | ScenarioList:
        if isinstance(external_id, str):
            scenario = self._retrieve((self.sources.space, external_id))

            extra_file_edges = self.extra_files.retrieve(external_id)
            scenario.extra_files = [edge.end_node.external_id for edge in extra_file_edges]
            mappings_override_edges = self.mappings_overrides.retrieve(external_id)
            scenario.mappings_override = [edge.end_node.external_id for edge in mappings_override_edges]

            return scenario
        else:
            scenarios = self._retrieve([(self.sources.space, ext_id) for ext_id in external_id])

            extra_file_edges = self.extra_files.retrieve(external_id)
            self._set_extra_files(scenarios, extra_file_edges)
            mappings_override_edges = self.mappings_overrides.retrieve(external_id)
            self._set_mappings_override(scenarios, mappings_override_edges)

            return scenarios

    def list(self, limit: int = INSTANCES_LIST_LIMIT_DEFAULT) -> ScenarioList:
        scenarios = self._list(limit=limit)

        extra_file_edges = self.extra_files.list(limit=-1)
        self._set_extra_files(scenarios, extra_file_edges)
        mappings_override_edges = self.mappings_overrides.list(limit=-1)
        self._set_mappings_override(scenarios, mappings_override_edges)

        return scenarios

    @staticmethod
    def _set_extra_files(scenarios: Sequence[Scenario], extra_file_edges: Sequence[dm.Edge]):
        edges_by_start_node: Dict[Tuple, List] = defaultdict(list)
        for edge in extra_file_edges:
            edges_by_start_node[edge.start_node.as_tuple()].append(edge)

        for scenario in scenarios:
            node_id = scenario.id_tuple()
            if node_id in edges_by_start_node:
                scenario.extra_files = [edge.end_node.external_id for edge in edges_by_start_node[node_id]]

    @staticmethod
    def _set_mappings_override(scenarios: Sequence[Scenario], mappings_override_edges: Sequence[dm.Edge]):
        edges_by_start_node: Dict[Tuple, List] = defaultdict(list)
        for edge in mappings_override_edges:
            edges_by_start_node[edge.start_node.as_tuple()].append(edge)

        for scenario in scenarios:
            node_id = scenario.id_tuple()
            if node_id in edges_by_start_node:
                scenario.mappings_override = [edge.end_node.external_id for edge in edges_by_start_node[node_id]]



from __future__ import annotations


from typing import ClassVar, Optional, Union

from pydantic import Field, field_validator

from cognite.powerops.cdf_labels import AssetLabel
from cognite.powerops.resync.models.cdf_resources import CDFSequence

from .base import Bid, Market, Process, ShopTransformation
from cognite.powerops.resync.utils.serializer import try_load_dict


class DayAheadBid(Bid):
    is_default_config_for_price_area: bool = True
    main_scenario: str
    price_area: str
    price_scenarios: dict[str, str]
    no_shop: bool = False
    bid_process_configuration_name: str
    bid_matrix_generator_config_external_id: str
    market_config_external_id: str

    @field_validator("price_scenarios", mode="before")
    def parse_str(cls, value) -> dict:
        value = try_load_dict(value)
        if isinstance(value, list) and not value:
            return {}
        return value


class DayAheadProcess(Process):
    parent_external_id = "bid_process_configurations"
    label: ClassVar[Union[AssetLabel, str]] = AssetLabel.BID_PROCESS_CONFIGURATION
    parent_description: ClassVar[str] = "Configurations used in bid matrix generation processes"
    shop: ShopTransformation
    bid: DayAheadBid
    bid_matrix_generator_config: Optional[CDFSequence] = None
    incremental_mapping: list[CDFSequence] = Field(default_factory=list)

    @field_validator("bid", mode="before")
    def parse_str(cls, value) -> dict:
        return try_load_dict(value)


class NordPoolMarket(Market):
    max_price: float
    min_price: float
    price_steps: int
    price_unit: str
    tick_size: float
    time_unit: str
    trade_lot: float



from __future__ import annotations

import json
import logging
from json import JSONDecodeError
from typing import Optional

import pandas as pd
from cognite.client import CogniteClient

from cognite.powerops.cdf_labels import RelationshipLabel
from cognite.powerops.clients.shop.api.shop_result_files_api import ShopFilesAPI
from cognite.powerops.clients.shop.data_classes import ShopRun
from cognite.powerops.clients.shop.data_classes.shop_result_files import ShopLogFile, ShopYamlFile
from cognite.powerops.clients.shop.data_classes.shop_results import ObjectiveFunction, ShopRunResult
from cognite.powerops.clients.shop.data_classes.shop_results_compare import ShopResultsCompare
from cognite.powerops.utils.cdf.calls import retrieve_relationships_from_source_ext_id

logger = logging.getLogger(__name__)


class ShopRunResultsAPI:
    def __init__(self, client: CogniteClient, files_api: ShopFilesAPI):
        self._client = client
        self._files_api = files_api
        self.compare = ShopResultsCompare()

    def retrieve(self, shop_run: ShopRun) -> ShopRunResult:
        if shop_run.in_progress:
            raise ValueError("ShopRun not completed.")

        post_run = None
        cplex = None
        shop_messages = None

        related_log_files = self._files_api.retrieve_related_files_metadata(
            source_external_id=shop_run.shop_run_event.external_id,
            label_ext_id=RelationshipLabel.LOG_FILE.value,
        )
        for metadata in related_log_files:
            ext_id = metadata.external_id
            if ext_id.endswith(".log") and "cplex" in ext_id:
                cplex: Optional[ShopLogFile] = self._files_api.log_files.retrieve(metadata)
            elif ext_id.endswith(".log") and "shop_messages" in ext_id:
                shop_messages: Optional[ShopLogFile] = self._files_api.log_files.retrieve(metadata)
            elif ext_id.endswith(".yaml"):
                post_run: Optional[ShopYamlFile] = self._files_api.yaml_files.retrieve(metadata)
            else:
                logger.error("Unknown file type")
        return ShopRunResult(self.retrieve_objective_function, shop_run, cplex, shop_messages, post_run)

    def retrieve_objective_function(self, shop_run: ShopRun) -> ObjectiveFunction:
        # TODO: ability to retrieve the objective function from the post run yaml file
        relationships = retrieve_relationships_from_source_ext_id(
            self._client,
            shop_run.shop_run_event.external_id,
            RelationshipLabel.OBJECTIVE_SEQUENCE,
            target_types=["sequence"],
        )
        sequences = self._client.sequences.retrieve_multiple(external_ids=[r.target_external_id for r in relationships])
        for seq in sequences:
            # sometimes the label is given to the non-objective sequence too
            if "objective" not in seq.name.lower():
                continue
            # Data is inserted as a single row DataFrame
            seq_data = self._client.sequences.data.retrieve_dataframe(
                external_id=seq.external_id, start=0, end=-1
            ).to_dict("records")[0]
            # This shape will always be (72, 7) 72 fields and 7 properties of each field
            column_definitions_df = pd.DataFrame(seq.columns)

            try:
                penalty_breakdown = json.loads(seq.metadata.get("shop:penalty_breakdown", {}))
                if type(penalty_breakdown) != dict:
                    penalty_breakdown = {}
            except JSONDecodeError:
                penalty_breakdown = {}

            return ObjectiveFunction(
                external_id=seq.external_id,
                data=seq_data,
                watercourse=seq.metadata.get("shop:watercourse", ""),
                penalty_breakdown=penalty_breakdown,
                field_definitions_df=column_definitions_df,
            )

        logger.error("Objective function sequence not found")
        raise ValueError("Objective function sequence not found")



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, ClassVar, Optional

import pandas as pd
from cognite.client import CogniteClient
from cognite.client.data_classes import FileMetadata, Sequence
from cognite.client.exceptions import CogniteNotFoundError
from pydantic import BaseModel, ConfigDict
from cognite.powerops.resync.utils.serializer import remove_read_only_fields


class _CDFResource(BaseModel, ABC):
    model_config: ClassVar[ConfigDict] = ConfigDict(arbitrary_types_allowed=True)

    @abstractmethod
    def _dump(self, camel_case: bool = False) -> dict[str, Any]:
        ...

    def dump(self, camel_case: bool = False) -> dict[str, Any]:
        dump = self._dump(camel_case)
        remove_read_only_fields(dump)
        return dump

    @classmethod
    @abstractmethod
    def from_cdf(
        cls,
        client: CogniteClient,
        resource_ext_id: str,
        fetch_content: bool = False,
    ) -> _CDFResource:
        ...


class CDFSequence(_CDFResource):
    sequence: Sequence
    content: Optional[pd.DataFrame] = None

    def __repr__(self) -> str:
        return f"CDFSequence(external_id={self.external_id})"

    def __str__(self) -> str:
        return self.__repr__()

    @property
    def external_id(self):
        return self.sequence.external_id

    def _set_data_set_id(self, data_set_id: int):
        self.sequence.data_set_id = data_set_id

    data_set_id = property(None, fset=_set_data_set_id)

    def _dump(self, camel_case: bool = False) -> dict[str, Any]:
        return self.sequence.dump(camel_case=camel_case)

    @classmethod
    def from_cdf(
        cls,
        client: CogniteClient,
        resource_ext_id: str,
        fetch_content: bool = False,
    ) -> CDFSequence:
        sequence = client.sequences.retrieve(external_id=resource_ext_id)
        if sequence is None:
            raise CogniteNotFoundError([resource_ext_id])
        if fetch_content:
            # limit defaults to 100, might not be an issue
            content = client.sequences.data.retrieve_dataframe(
                external_id=resource_ext_id,
                start=0,
                end=None,
            )
        else:
            content = None
        return cls(sequence=sequence, content=content)


class CDFFile(_CDFResource):
    meta: FileMetadata
    content: Optional[bytes] = None

    def __repr__(self) -> str:
        return f"CDFFile(external_id={self.external_id})"

    def __str__(self) -> str:
        return self.__repr__()

    @property
    def external_id(self):
        return self.meta.external_id

    def _dump(self, camel_case: bool = False) -> dict[str, Any]:
        return self.meta.dump(camel_case=camel_case)

    @classmethod
    def from_cdf(
        cls,
        client: CogniteClient,
        resource_ext_id: str,
        fetch_content: bool = False,
    ) -> CDFFile:
        meta = client.files.retrieve(external_id=resource_ext_id)
        if meta is None:
            raise CogniteNotFoundError([resource_ext_id])
        if fetch_content:
            content = client.files.download_bytes(external_id=resource_ext_id)
        else:
            content = None
        return cls(meta=meta, content=content)
